from __future__ import annotations

import asyncio
import os
from asyncio import Event
from typing import MutableSequence, TYPE_CHECKING, Tuple, Type, Union

import pkg_resources

from streamflow.core.deployment import Connector, ConnectorCopyKind, DeploymentManager
from streamflow.core.scheduling import Location
from streamflow.deployment.connector import connector_classes
from streamflow.log_handler import logger

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from streamflow.core.deployment import DeploymentConfig
    from typing import MutableMapping, Optional, Any


class DefaultDeploymentManager(DeploymentManager):

    def __init__(self,
                 context: StreamFlowContext) -> None:
        super().__init__(context)
        self.config_map: MutableMapping[str, Any] = {}
        self.events_map: MutableMapping[str, Event] = {}
        self.deployments_map: MutableMapping[str, Connector] = {}

    async def close(self):
        await self.undeploy_all()

    async def deploy(self, deployment_config: DeploymentConfig):
        deployment_name = deployment_config.name
        while True:
            if deployment_name not in self.events_map:
                self.events_map[deployment_name] = Event()
            if deployment_name not in self.config_map:
                self.config_map[deployment_name] = deployment_config
                if deployment_config.lazy:
                    connector = FutureConnector(
                        name=deployment_name,
                        context=self.context,
                        type=connector_classes[deployment_config.type],
                        external=deployment_config.external,
                        **deployment_config.config)
                    self.deployments_map[deployment_name] = connector
                    self.events_map[deployment_name].set()
                else:
                    connector = connector_classes[deployment_config.type](
                        deployment_name, self.context, **deployment_config.config)
                    self.deployments_map[deployment_name] = connector
                    if not deployment_config.external:
                        logger.info("Deploying {}".format(deployment_name))
                    await connector.deploy(deployment_config.external)
                    if not deployment_config.external:
                        logger.info("Deployment of {} terminated with status COMPLETED".format(deployment_name))
                    self.events_map[deployment_name].set()
                    break
            else:
                await self.events_map[deployment_name].wait()
                if deployment_name in self.config_map:
                    break

    def get_connector(self, deployment_name: str) -> Optional[Connector]:
        return self.deployments_map.get(deployment_name, None)

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'deployment_manager.json'))

    def is_deployed(self, deployment_name: str):
        return deployment_name in self.deployments_map

    async def undeploy(self, deployment_name: str):
        if deployment_name in dict(self.deployments_map):
            await self.events_map[deployment_name].wait()
            self.events_map[deployment_name].clear()
            connector = self.deployments_map[deployment_name]
            config = self.config_map[deployment_name]
            if not config.external:
                logger.info("Undeploying {deployment}".format(deployment=deployment_name))
            await connector.undeploy(config.external)
            if not config.external:
                logger.info("Undeployment of {} terminated with status COMPLETED".format(deployment_name))
            del self.deployments_map[deployment_name]
            del self.config_map[deployment_name]
            self.events_map[deployment_name].set()

    async def undeploy_all(self):
        undeployments = []
        for name in dict(self.deployments_map):
            undeployments.append(asyncio.create_task(self.undeploy(name)))
        await asyncio.gather(*undeployments)


class FutureConnector(Connector):

    def __init__(self,
                 name: str,
                 context: StreamFlowContext,
                 type: Type[Connector],
                 external: bool,
                 **kwargs):
        super().__init__(name, context)
        self.type: Type[Connector] = type
        self.external: bool = external
        self.parameters: MutableMapping[str, Any] = kwargs
        self.deploying: bool = False
        self.deploy_event: Event = Event()
        self.connector: Optional[Connector] = None

    async def copy(self,
                   src: str,
                   dst: str,
                   locations: MutableSequence[str],
                   kind: ConnectorCopyKind,
                   source_connector: Optional[Connector] = None,
                   source_location: Optional[str] = None,
                   read_only: bool = False) -> None:
        if self.connector is None:
            if not self.deploying:
                self.deploying = True
                await self.deploy(self.external)
            else:
                await self.deploy_event.wait()
        if isinstance(source_connector, FutureConnector):
            source_connector = source_connector.connector
        await self.connector.copy(
            src=src,
            dst=dst,
            locations=locations,
            kind=kind,
            source_connector=source_connector,
            source_location=source_location,
            read_only=read_only)

    async def deploy(self,
                     external: bool) -> None:
        # noinspection PyArgumentList
        connector = self.type(
            self.deployment_name, self.context, **self.parameters)
        if not external:
            logger.info("Deploying {}".format(self.deployment_name))
        await connector.deploy(external)
        if not external:
            logger.info("Deployment of {} terminated with status COMPLETED".format(self.deployment_name))
        self.connector = connector
        self.deploy_event.set()

    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        if self.connector is None:
            if not self.deploying:
                self.deploying = True
                await self.deploy(self.external)
            else:
                await self.deploy_event.wait()
        return await self.connector.get_available_locations(
            service=service,
            input_directory=input_directory,
            output_directory=output_directory,
            tmp_directory=tmp_directory)

    def get_schema(self) -> str:
        return self.type.get_schema()

    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        if self.connector is None:
            if not self.deploying:
                self.deploying = True
                await self.deploy(self.external)
            else:
                await self.deploy_event.wait()
        return await self.connector.run(
            location=location,
            command=command,
            environment=environment,
            workdir=workdir,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            capture_output=capture_output,
            job_name=job_name)

    async def undeploy(self, external: bool) -> None:
        if self.connector is not None:
            await self.connector.undeploy(external)



from __future__ import annotations

import math
from typing import Any, MutableMapping, MutableSequence, Optional, Union

from streamflow.core.context import StreamFlowContext
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.scheduling import Hardware, HardwareRequirement
from streamflow.core.utils import get_token_value
from streamflow.core.workflow import Token
from streamflow.cwl.utils import eval_expression


class CWLHardwareRequirement(HardwareRequirement):

    def __init__(self,
                 cores: Union[str, float] = 1,
                 memory: Union[str, float] = 256,
                 tmpdir: Union[str, float] = 1024,
                 outdir: Union[str, float] = 1024,
                 full_js: bool = False,
                 expression_lib: Optional[MutableSequence[str]] = None, ):
        self.cores: Union[str, float] = cores
        self.memory: Union[str, float] = memory
        self.tmpdir: Union[str, float] = tmpdir
        self.outdir: Union[str, float] = outdir
        self.full_js: bool = full_js
        self.expression_lib: Optional[MutableSequence[str]] = expression_lib

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CWLHardwareRequirement:
        return cls(
            cores=row['cores'],
            memory=row['memory'],
            tmpdir=row['tmpdir'],
            outdir=row['outdir'],
            full_js=row['full_js'],
            expression_lib=row['expression_lib'])

    async def _save_additional_params(self, context: StreamFlowContext):
        return {
            'cores': self.cores,
            'memory': self.memory,
            'tmpdir': self.tmpdir,
            'outdir': self.outdir,
            'full_js': self.full_js,
            'expression_lib': self.expression_lib}

    def _process_requirement(self,
                             requirement: Union[str, float],
                             context: MutableMapping[str, Any]) -> float:
        return math.ceil(eval_expression(
            expression=requirement,
            context=context,
            full_js=self.full_js,
            expression_lib=self.expression_lib))

    def eval(self, inputs: MutableMapping[str, Token]) -> Hardware:
        context = {'inputs': {name: get_token_value(t) for name, t in inputs.items()}}
        return Hardware(
            cores=self._process_requirement(self.cores, context),
            memory=self._process_requirement(self.memory, context),
            tmp_directory=self._process_requirement(self.tmpdir, context),
            output_directory=self._process_requirement(self.outdir, context))



from __future__ import annotations

import copy
import os
import posixpath
import tempfile
import urllib.parse
from enum import Enum
from pathlib import PurePosixPath
from typing import Any, MutableMapping, MutableSequence, Optional, Set, Type, Union, cast

import cwltool.command_line_tool
import cwltool.context
import cwltool.docker_id
import cwltool.process
import cwltool.workflow
from rdflib import Graph
from ruamel.yaml.comments import CommentedSeq

from streamflow.config.config import WorkflowConfig
from streamflow.core.context import StreamFlowContext
from streamflow.core.data import LOCAL_LOCATION
from streamflow.core.deployment import DeploymentConfig, LocalTarget, Target
from streamflow.core.exception import WorkflowDefinitionException
from streamflow.core.utils import get_tag, random_name
from streamflow.core.workflow import CommandOutputProcessor, Port, Step, Token, TokenProcessor, Workflow
from streamflow.cwl import utils
from streamflow.cwl.combinator import ListMergeCombinator
from streamflow.cwl.command import (
    CWLCommand, CWLCommandToken, CWLExpressionCommand, CWLMapCommandToken, CWLObjectCommandToken, CWLUnionCommandToken
)
from streamflow.cwl.hardware import CWLHardwareRequirement
from streamflow.cwl.processor import (
    CWLCommandOutputProcessor, CWLMapCommandOutputProcessor, CWLMapTokenProcessor, CWLObjectCommandOutputProcessor,
    CWLObjectTokenProcessor, CWLTokenProcessor, CWLUnionCommandOutputProcessor, CWLUnionTokenProcessor
)
from streamflow.cwl.step import (
    CWLConditionalStep, CWLEmptyScatterConditionalStep, CWLInputInjectorStep,
    CWLLoopConditionalStep, CWLLoopOutputAllStep, CWLLoopOutputLastStep, CWLTransferStep
)
from streamflow.cwl.transformer import (
    AllNonNullTransformer, CWLDefaultTransformer, CWLTokenTransformer, FirstNonNullTransformer, ForwardTransformer,
    ListToElementTransformer, LoopValueFromTransformer, OnlyNonNullTransformer, ValueFromTransformer
)
from streamflow.cwl.utils import LoadListing, SecondaryFile, resolve_dependencies
from streamflow.log_handler import logger
from streamflow.workflow.combinator import (
    CartesianProductCombinator, DotProductCombinator, LoopCombinator,
    LoopTerminationCombinator
)
from streamflow.workflow.step import (
    Combinator, CombinatorStep, DeployStep, ExecuteStep, GatherStep, LoopCombinatorStep, ScatterStep, ScheduleStep,
    Transformer
)
from streamflow.workflow.token import TerminationToken


def _create_command(
        cwl_element: cwltool.command_line_tool.CommandLineTool,
        cwl_name_prefix: str,
        schema_def_types: MutableMapping[str, Any],
        context: MutableMapping[str, Any],
        step: Step) -> CWLCommand:
    command = CWLCommand(step)
    # Process InitialWorkDirRequirement
    requirements = {**context['hints'], **context['requirements']}
    if 'InitialWorkDirRequirement' in requirements:
        command.initial_work_dir = requirements['InitialWorkDirRequirement']['listing']
        command.absolute_initial_workdir_allowed = 'DockerRequirement' in context['requirements']
    # Process InplaceUpdateRequirement
    if 'InplaceUpdateRequirement' in requirements:
        command.inplace_update = requirements['InplaceUpdateRequirement']['inplaceUpdate']
    # Process EnvVarRequirement
    if 'EnvVarRequirement' in requirements:
        for env_entry in requirements['EnvVarRequirement']['envDef']:
            command.environment[env_entry['envName']] = env_entry['envValue']
    # Process ShellCommandRequirement
    if 'ShellCommandRequirement' in requirements:
        command.is_shell_command = True
    # Process success and failure codes
    if 'successCodes' in cwl_element.tool:
        command.success_codes = cwl_element.tool['successCodes']
    if 'permanentFailCodes' or 'temporaryFailCodes' in cwl_element.tool:
        command.failure_codes = cwl_element.tool.get('permanentFailCodes', [])
        command.failure_codes.extend(cwl_element.tool.get('temporaryFailCodes', []))
    # Process baseCommand
    if 'baseCommand' in cwl_element.tool:
        if isinstance(cwl_element.tool['baseCommand'], CommentedSeq):
            for command_token in cwl_element.tool['baseCommand']:
                command.base_command.append(command_token)
        else:
            command.base_command.append(cwl_element.tool['baseCommand'])
    # Process arguments
    if 'arguments' in cwl_element.tool:
        for argument in cwl_element.tool['arguments']:
            command.command_tokens.append(_get_command_token(
                binding=argument,
                is_shell_command=command.is_shell_command))
    # Process inputs
    for input_port in cwl_element.tool['inputs']:
        command_token = _get_command_token_from_input(
            cwl_element=input_port,
            port_type=input_port['type'],
            input_name=_get_name('', cwl_name_prefix, input_port['id']),
            is_shell_command=command.is_shell_command,
            schema_def_types=schema_def_types)
        if command_token is not None:
            command.command_tokens.append(command_token)
    return command


def _create_command_output_processor(port_name: str,
                                     workflow: Workflow,
                                     port_target: Optional[Target],
                                     port_type: Any,
                                     cwl_element: MutableMapping[str, Any],
                                     cwl_name_prefix: str,
                                     schema_def_types: MutableMapping[str, Any],
                                     context: MutableMapping[str, Any],
                                     optional: bool = False) -> CommandOutputProcessor:
    if isinstance(port_type, MutableMapping):
        if 'type' in port_type:
            # Array type: -> MapCommandOutputProcessor
            if port_type['type'] == 'array':
                return CWLMapCommandOutputProcessor(
                    name=port_name,
                    workflow=workflow,
                    processor=_create_command_output_processor(
                        port_name=port_name,
                        workflow=workflow,
                        port_target=port_target,
                        port_type=port_type['items'],
                        cwl_element=cwl_element,
                        cwl_name_prefix=cwl_name_prefix,
                        schema_def_types=schema_def_types,
                        context=context,
                        optional=optional))
            # Enum type: -> create command output processor
            elif port_type['type'] == 'enum':
                # Process InlineJavascriptRequirement
                requirements = {**context['hints'], **context['requirements']}
                expression_lib, full_js = _process_javascript_requirement(requirements)
                enum_prefix = (_get_name(posixpath.sep, posixpath.sep, port_type['name']) if 'name' in port_type
                               else cwl_name_prefix)
                # Return OutputProcessor
                return CWLCommandOutputProcessor(
                    name=port_name,
                    workflow=workflow,
                    target=port_target,
                    token_type=port_type['type'],
                    enum_symbols=[posixpath.relpath(_get_name(posixpath.sep, posixpath.sep, s), enum_prefix)
                                  for s in port_type['symbols']],
                    expression_lib=expression_lib,
                    full_js=full_js,
                    optional=optional)
            # Record type: -> ObjectCommandOutputProcessor
            elif port_type['type'] == 'record':
                # Process InlineJavascriptRequirement
                requirements = {**context['hints'], **context['requirements']}
                expression_lib, full_js = _process_javascript_requirement(requirements)
                # Create processor
                record_name_prefix = _get_name(posixpath.sep, posixpath.sep, port_type.get('name', port_name))
                return CWLObjectCommandOutputProcessor(
                    name=port_name,
                    workflow=workflow,
                    processors={_get_name('', record_name_prefix, port_type['name']): _create_command_output_processor(
                        port_name=port_name,
                        port_target=port_target,
                        workflow=workflow,
                        port_type=port_type['type'],
                        cwl_element=port_type,
                        cwl_name_prefix=posixpath.join(record_name_prefix,
                                                       _get_name('', record_name_prefix, port_type['name'])),
                        schema_def_types=schema_def_types,
                        context=context) for port_type in port_type['fields']},
                    expression_lib=expression_lib,
                    full_js=full_js,
                    output_eval=cwl_element.get('outputBinding', {}).get('outputEval'))
            # Unknown type -> raise Exception
            else:
                raise WorkflowDefinitionException("Unsupported type {} for port {}.".format(
                    port_type['type'], port_name))
        # Untyped object -> not supported
        else:
            raise WorkflowDefinitionException("Unsupported dictionary type without explicit `type` key")
    elif isinstance(port_type, MutableSequence):
        optional = 'null' in port_type
        types = [t for t in filter(lambda x: x != 'null', port_type)]
        # Optional type (e.g. ['null', Type] -> Equivalent to Type?
        if len(types) == 1:
            return _create_command_output_processor(
                port_name=port_name,
                workflow=workflow,
                port_target=port_target,
                port_type=types[0],
                cwl_element=cwl_element,
                cwl_name_prefix=cwl_name_prefix,
                schema_def_types=schema_def_types,
                context=context,
                optional=optional)
        # List of types: -> UnionOutputProcessor
        else:
            return CWLUnionCommandOutputProcessor(
                name=port_name,
                workflow=workflow,
                processors=[_create_command_output_processor(
                    port_name=port_name,
                    workflow=workflow,
                    port_target=port_target,
                    port_type=port_type,
                    cwl_element=cwl_element,
                    cwl_name_prefix=cwl_name_prefix,
                    schema_def_types=schema_def_types,
                    context=context,
                    optional=optional) for port_type in types])
    # Complex type -> Extract from schema definitions and propagate
    elif '#' in port_type:
        return _create_command_output_processor(
            port_name=port_name,
            workflow=workflow,
            port_target=port_target,
            port_type=schema_def_types[port_type],
            cwl_element=cwl_element,
            cwl_name_prefix=cwl_name_prefix,
            schema_def_types=schema_def_types,
            context=context,
            optional=optional)
    # Simple type -> Create typed processor
    else:
        # Process InlineJavascriptRequirement
        requirements = {**context['hints'], **context['requirements']}
        expression_lib, full_js = _process_javascript_requirement(requirements)
        # Create OutputProcessor
        if port_type == 'File':
            return CWLCommandOutputProcessor(
                name=port_name,
                workflow=workflow,
                target=port_target,
                token_type=port_type,
                expression_lib=expression_lib,
                file_format=cwl_element.get('format'),
                full_js=full_js,
                glob=cwl_element.get('outputBinding', {}).get('glob', cwl_element.get('path')),
                load_contents=cwl_element.get(
                    'loadContents', cwl_element.get('outputBinding', {}).get('loadContents', False)),
                load_listing=_get_load_listing(cwl_element, context),
                optional=optional,
                output_eval=cwl_element.get('outputBinding', {}).get('outputEval'),
                secondary_files=_get_secondary_files(cwl_element.get('secondaryFiles'), default_required=False),
                streamable=cwl_element.get('streamable', False))
        else:
            # Normalize port type (Python does not distinguish among all CWL number types)
            port_type = 'long' if port_type == 'int' else 'double' if port_type == 'float' else port_type
            return CWLCommandOutputProcessor(
                name=port_name,
                workflow=workflow,
                target=port_target,
                token_type=port_type,
                expression_lib=expression_lib,
                full_js=full_js,
                glob=cwl_element.get('outputBinding', {}).get('glob', cwl_element.get('path')),
                load_contents=cwl_element.get(
                    'loadContents', cwl_element.get('outputBinding', {}).get('loadContents', False)),
                load_listing=_get_load_listing(cwl_element, context),
                optional=optional,
                output_eval=cwl_element.get('outputBinding', {}).get('outputEval'))


def _create_context() -> MutableMapping[str, Any]:
    return {
        'default': {},
        'requirements': {},
        'hints': {}
    }


def _create_list_merger(name: str,
                        workflow: Workflow,
                        ports: MutableMapping[str, Port],
                        output_port: Optional[Port] = None,
                        link_merge: Optional[str] = None,
                        pick_value: Optional[str] = None) -> Step:
    combinator = workflow.create_step(
        cls=CombinatorStep,
        name=name + "-combinator",
        combinator=ListMergeCombinator(
            name=utils.random_name(),
            workflow=workflow,
            input_names=list(ports.keys()),
            output_name=name,
            flatten=(link_merge == 'merge_flattened')))
    for port_name, port in ports.items():
        combinator.add_input_port(port_name, port)
        combinator.combinator.add_item(port_name)
    if pick_value == 'first_non_null':
        combinator.add_output_port(name, workflow.create_port())
        transformer = workflow.create_step(
            cls=FirstNonNullTransformer,
            name=name + "-transformer")
        transformer.add_input_port(name, combinator.get_output_port())
        transformer.add_output_port(name, output_port or workflow.create_port())
        return transformer
    elif pick_value == 'the_only_non_null':
        combinator.add_output_port(name, workflow.create_port())
        transformer = workflow.create_step(
            cls=OnlyNonNullTransformer,
            name=name + "-transformer")
        transformer.add_input_port(name, combinator.get_output_port())
        transformer.add_output_port(name, output_port or workflow.create_port())
        return transformer
    elif pick_value == 'all_non_null':
        combinator.add_output_port(name, workflow.create_port())
        transformer = workflow.create_step(
            cls=AllNonNullTransformer,
            name=name + "-transformer")
        if link_merge is None:
            list_to_element = workflow.create_step(
                cls=ListToElementTransformer,
                name=name + "-list-to-element")
            list_to_element.add_input_port(name, combinator.get_output_port())
            list_to_element.add_output_port(name, workflow.create_port())
            transformer.add_input_port(name, list_to_element.get_output_port())
            transformer.add_output_port(name, output_port or workflow.create_port())
        else:
            transformer.add_input_port(name, combinator.get_output_port())
            transformer.add_output_port(name, output_port or workflow.create_port())
        return transformer
    elif link_merge is None:
        combinator.add_output_port(name, workflow.create_port())
        list_to_element = workflow.create_step(
            cls=ListToElementTransformer,
            name=name + "-list-to-element")
        list_to_element.add_input_port(name, combinator.get_output_port())
        list_to_element.add_output_port(name, output_port or workflow.create_port())
        return list_to_element
    else:
        combinator.add_output_port(name, output_port or workflow.create_port())
        return combinator


def _create_residual_combinator(workflow: Workflow,
                                step_name: str,
                                inner_combinator: Combinator,
                                inner_inputs: MutableSequence[str],
                                input_ports: MutableMapping[str, Port]) -> Combinator:
    dot_product_combinator = DotProductCombinator(
        workflow=workflow,
        name=step_name + "-dot-product-combinator")
    if inner_combinator:
        dot_product_combinator.add_combinator(
            inner_combinator, inner_combinator.get_items(recursive=True))
    else:
        for global_name in inner_inputs:
            dot_product_combinator.add_item(posixpath.relpath(global_name, step_name))
    for global_name in input_ports:
        if global_name not in inner_inputs:
            dot_product_combinator.add_item(posixpath.relpath(global_name, step_name))
    return dot_product_combinator


def _create_token_processor(port_name: str,
                            workflow: Workflow,
                            port_type: Any,
                            cwl_element: MutableMapping[str, Any],
                            cwl_name_prefix: str,
                            schema_def_types: MutableMapping[str, Any],
                            format_graph: Graph,
                            context: MutableMapping[str, Any],
                            optional: bool = False,
                            default_required_sf: bool = True,
                            check_type: bool = True,
                            force_deep_listing: bool = False,
                            only_propagate_secondary_files: bool = True) -> TokenProcessor:
    if isinstance(port_type, MutableMapping):
        if 'type' in port_type:
            # Array type: -> MapTokenProcessor
            if port_type['type'] == 'array':
                return CWLMapTokenProcessor(
                    name=port_name,
                    workflow=workflow,
                    processor=_create_token_processor(
                        port_name=port_name,
                        workflow=workflow,
                        port_type=port_type['items'],
                        cwl_element=cwl_element,
                        cwl_name_prefix=cwl_name_prefix,
                        schema_def_types=schema_def_types,
                        format_graph=format_graph,
                        context=context,
                        optional=optional,
                        check_type=check_type,
                        force_deep_listing=force_deep_listing,
                        only_propagate_secondary_files=only_propagate_secondary_files))
            # Enum type: -> create output processor
            elif port_type['type'] == 'enum':
                # Process InlineJavascriptRequirement
                requirements = {**context['hints'], **context['requirements']}
                expression_lib, full_js = _process_javascript_requirement(requirements)
                enum_prefix = (_get_name(posixpath.sep, posixpath.sep, port_type['name']) if 'name' in port_type
                               else cwl_name_prefix)
                # Return TokenProcessor
                return CWLTokenProcessor(
                    name=port_name,
                    workflow=workflow,
                    token_type=port_type['type'],
                    check_type=check_type,
                    enum_symbols=[posixpath.relpath(_get_name(posixpath.sep, posixpath.sep, s), enum_prefix)
                                  for s in port_type['symbols']],
                    expression_lib=expression_lib,
                    full_js=full_js,
                    optional=optional)
            # Record type: -> ObjectTokenProcessor
            elif port_type['type'] == 'record':
                record_name_prefix = _get_name(posixpath.sep, posixpath.sep, port_type.get('name', port_name))
                return CWLObjectTokenProcessor(
                    name=port_name,
                    workflow=workflow,
                    processors={_get_name('', record_name_prefix, port_type['name']): _create_token_processor(
                        port_name=port_name,
                        workflow=workflow,
                        port_type=port_type['type'],
                        cwl_element=port_type,
                        cwl_name_prefix=posixpath.join(record_name_prefix,
                                                       _get_name('', record_name_prefix, port_type['name'])),
                        schema_def_types=schema_def_types,
                        format_graph=format_graph,
                        context=context,
                        check_type=check_type,
                        force_deep_listing=force_deep_listing,
                        only_propagate_secondary_files=only_propagate_secondary_files
                    ) for port_type in port_type['fields']})
            # Unknown type -> raise Exception
            else:
                raise WorkflowDefinitionException("Unsupported type {} for port {}.".format(
                    port_type['type'], port_name))
        # Untyped object -> not supported
        else:
            raise WorkflowDefinitionException("Unsupported dictionary type without explicit `type` key")
    elif isinstance(port_type, MutableSequence):
        optional = 'null' in port_type
        types = [t for t in filter(lambda x: x != 'null', port_type)]
        # Optional type (e.g. ['null', Type] -> Equivalent to Type?
        if len(types) == 1:
            return _create_token_processor(
                port_name=port_name,
                workflow=workflow,
                port_type=types[0],
                cwl_element=cwl_element,
                cwl_name_prefix=cwl_name_prefix,
                schema_def_types=schema_def_types,
                format_graph=format_graph,
                context=context,
                optional=optional,
                check_type=check_type,
                force_deep_listing=force_deep_listing,
                only_propagate_secondary_files=only_propagate_secondary_files)
        # List of types: -> UnionOutputProcessor
        else:
            return CWLUnionTokenProcessor(
                name=port_name,
                workflow=workflow,
                processors=[_create_token_processor(
                    port_name=port_name,
                    workflow=workflow,
                    port_type=port_type,
                    cwl_element=cwl_element,
                    cwl_name_prefix=cwl_name_prefix,
                    schema_def_types=schema_def_types,
                    format_graph=format_graph,
                    context=context,
                    optional=optional,
                    check_type=check_type,
                    force_deep_listing=force_deep_listing,
                    only_propagate_secondary_files=only_propagate_secondary_files
                ) for port_type in types])
    # Complex type -> Extract from schema definitions and propagate
    elif '#' in port_type:
        return _create_token_processor(
            port_name=port_name,
            workflow=workflow,
            port_type=schema_def_types[port_type],
            cwl_element=cwl_element,
            cwl_name_prefix=cwl_name_prefix,
            schema_def_types=schema_def_types,
            format_graph=format_graph,
            context=context,
            optional=optional,
            check_type=check_type,
            force_deep_listing=force_deep_listing,
            only_propagate_secondary_files=only_propagate_secondary_files)
    # Simple type -> Create typed processor
    else:
        # Process InlineJavascriptRequirement
        requirements = {**context['hints'], **context['requirements']}
        expression_lib, full_js = _process_javascript_requirement(requirements)
        # Create OutputProcessor
        if port_type == 'File':
            return CWLTokenProcessor(
                name=port_name,
                workflow=workflow,
                token_type=port_type,
                check_type=check_type,
                expression_lib=expression_lib,
                file_format=cwl_element.get('format'),
                format_graph=format_graph,
                full_js=full_js,
                load_contents=cwl_element.get(
                    'loadContents', cwl_element.get('inputBinding', {}).get('loadContents')),
                load_listing=(LoadListing.deep_listing if force_deep_listing else
                              _get_load_listing(cwl_element, context)),
                only_propagate_secondary_files=only_propagate_secondary_files,
                optional=optional,
                secondary_files=_get_secondary_files(
                    cwl_element=cwl_element.get('secondaryFiles'),
                    default_required=default_required_sf),
                streamable=cwl_element.get('streamable', False))
        else:
            # Normalize port type (Python does not distinguish among all CWL number types)
            port_type = 'long' if port_type == 'int' else 'double' if port_type == 'float' else port_type
            return CWLTokenProcessor(
                name=port_name,
                workflow=workflow,
                token_type=port_type,
                check_type=check_type,
                expression_lib=expression_lib,
                full_js=full_js,
                load_contents=cwl_element.get(
                    'loadContents', cwl_element.get('inputBinding', {}).get('loadContents')),
                load_listing=(LoadListing.deep_listing if force_deep_listing else
                              _get_load_listing(cwl_element, context)),
                optional=optional)


def _create_token_transformer(name: str,
                              port_name: str,
                              workflow: Workflow,
                              cwl_element: MutableMapping[str, Any],
                              cwl_name_prefix: str,
                              schema_def_types: MutableMapping[str, Any],
                              format_graph: Graph,
                              context: MutableMapping[str, Any],
                              check_type: bool = True,
                              only_propagate_secondary_files: bool = True) -> CWLTokenTransformer:
    token_transformer = workflow.create_step(
        cls=CWLTokenTransformer,
        name=name,
        port_name=port_name,
        processor=_create_token_processor(
            port_name=port_name,
            workflow=workflow,
            port_type=cwl_element['type'],
            cwl_element=cwl_element,
            cwl_name_prefix=cwl_name_prefix,
            schema_def_types=schema_def_types,
            format_graph=format_graph,
            context=context,
            check_type=check_type,
            only_propagate_secondary_files=only_propagate_secondary_files))
    token_transformer.add_output_port(port_name, workflow.create_port())
    return token_transformer


def _dict_to_lists(groups: MutableMapping[str, Any], port_name: str) -> Token:
    if isinstance(groups, MutableMapping):
        groups_list = [_dict_to_lists(g, port_name) for g in groups.values()]
    else:
        groups_list = groups
    return Token(
        tag='.'.join(get_tag(groups_list).split('.')[:-1]),
        value=sorted(groups_list, key=lambda t: int(t.tag.split('.')[-1])))


def _get_command_token(binding: Any,
                       is_shell_command: bool = False,
                       input_name: Optional[str] = None,
                       token_type: Optional[str] = None) -> CWLCommandToken:
    # Normalize type (Python does not distinguish among all CWL number types)
    token_type = 'long' if token_type == 'int' else 'double' if token_type == 'float' else token_type
    if isinstance(binding, MutableMapping):
        item_separator = binding.get('itemSeparator', None)
        position = binding.get('position', 0)
        prefix = binding.get('prefix', None)
        separate = binding.get('separate', True)
        shell_quote = binding.get('shellQuote', True)
        value = binding['valueFrom'] if 'valueFrom' in binding else None
        return CWLCommandToken(
            name=input_name,
            value=value,
            token_type=token_type,
            is_shell_command=is_shell_command,
            item_separator=item_separator,
            position=position,
            prefix=prefix,
            separate=separate,
            shell_quote=shell_quote)
    else:
        return CWLCommandToken(
            name=input_name,
            value=binding,
            token_type=token_type)


def _get_command_token_from_input(cwl_element: Any,
                                  port_type: Any,
                                  input_name: str,
                                  is_shell_command: bool = False,
                                  schema_def_types: Optional[MutableMapping[str, Any]] = None):
    token = None
    command_line_binding = cwl_element.get('inputBinding', None)
    if isinstance(port_type, MutableMapping):
        if 'type' in port_type:
            # Array type: -> CWLMapCommandToken
            if port_type['type'] == 'array':
                token = _get_command_token_from_input(
                    cwl_element=port_type,
                    port_type=port_type['items'],
                    input_name=input_name,
                    is_shell_command=is_shell_command,
                    schema_def_types=schema_def_types)
                if token is not None:
                    token = CWLMapCommandToken(
                        name=input_name,
                        value=token,
                        is_shell_command=is_shell_command)
            # Enum type: -> substitute the type with string and reprocess
            elif port_type['type'] == 'enum':
                return _get_command_token_from_input(
                    cwl_element=cwl_element,
                    port_type='string',
                    input_name=input_name,
                    is_shell_command=is_shell_command,
                    schema_def_types=schema_def_types)
            # Object type: -> CWLObjectCommandToken
            elif port_type['type'] == 'record':
                record_name_prefix = _get_name(posixpath.sep, posixpath.sep, port_type.get('name', input_name))
                command_tokens = {}
                for el in port_type['fields']:
                    key = _get_name('', record_name_prefix, el['name'])
                    el_token = _get_command_token_from_input(
                        cwl_element=el,
                        port_type=el['type'],
                        input_name=key,
                        is_shell_command=is_shell_command,
                        schema_def_types=schema_def_types)
                    if el_token is not None:
                        command_tokens[key] = el_token
                if command_tokens:
                    token = CWLObjectCommandToken(
                        name=input_name,
                        value=command_tokens,
                        is_shell_command=True,
                        shell_quote=False)
            # Unknown type -> raise Exception
            else:
                raise WorkflowDefinitionException("Unsupported type {} for port {}.".format(
                    port_type['type'], input_name))
        # Untyped object -> not supported
        else:
            raise WorkflowDefinitionException("Unsupported dictionary type without explicit `type` key")
    elif isinstance(port_type, MutableSequence):
        types = [t for t in filter(lambda x: x != 'null', port_type)]
        # Optional type (e.g. ['null', Type] -> propagate
        if len(types) == 1:
            return _get_command_token_from_input(
                cwl_element=cwl_element,
                port_type=types[0],
                input_name=input_name,
                is_shell_command=is_shell_command,
                schema_def_types=schema_def_types)
        # List of types: -> CWLUnionCommandToken
        else:
            command_tokens = []
            for i, port_type in enumerate(types):
                token = _get_command_token_from_input(
                    cwl_element=cwl_element,
                    port_type=port_type,
                    input_name=input_name,
                    is_shell_command=is_shell_command,
                    schema_def_types=schema_def_types)
                if token is not None:
                    command_tokens.append(token)
            if command_tokens:
                token = CWLUnionCommandToken(
                    name=input_name,
                    value=command_tokens,
                    is_shell_command=True,
                    shell_quote=False)
    elif isinstance(port_type, str):
        # Complex type -> Extract from schema definitions and propagate
        if '#' in port_type:
            return _get_command_token_from_input(
                cwl_element=cwl_element,
                port_type=schema_def_types[port_type],
                input_name=input_name,
                is_shell_command=is_shell_command,
                schema_def_types=schema_def_types)
    # Simple type with `inputBinding` specified -> CWLCommandToken
    if command_line_binding is not None:
        if token is not None:
            # By default, do not escape composite command tokens
            if 'shellQuote' not in command_line_binding:
                command_line_binding['shellQuote'] = False
                return _get_command_token(
                    binding={**command_line_binding, **{'valueFrom': token}},
                    is_shell_command=True,
                    input_name=input_name)
            else:
                return _get_command_token(
                    binding={**command_line_binding, **{'valueFrom': token}},
                    is_shell_command=is_shell_command,
                    input_name=input_name)
        else:
            return _get_command_token(
                binding=command_line_binding,
                is_shell_command=is_shell_command,
                input_name=input_name,
                token_type=port_type)
    # Simple type without `inputBinding` specified -> token
    else:
        return token


def _get_hardware_requirement(
        requirements: MutableMapping[str, Any],
        expression_lib: Optional[MutableSequence[str]],
        full_js: bool):
    hardware_requirement = CWLHardwareRequirement(
        expression_lib=expression_lib,
        full_js=full_js)
    if 'ResourceRequirement' in requirements:
        resource_requirement = requirements['ResourceRequirement']
        hardware_requirement.cores = resource_requirement.get(
            'coresMin', resource_requirement.get('coresMax', hardware_requirement.cores))
        hardware_requirement.memory = resource_requirement.get(
            'ramMin', resource_requirement.get('ramMax', hardware_requirement.memory))
        hardware_requirement.tmpdir = resource_requirement.get(
            'tmpdirMin', resource_requirement.get('tmpdirMax', hardware_requirement.tmpdir))
        hardware_requirement.outdir = resource_requirement.get(
            'outdirMin', resource_requirement.get('outdirMax', hardware_requirement.outdir))
    return hardware_requirement


def _get_load_listing(port_description: MutableMapping[str, Any],
                      context: MutableMapping[str, Any]) -> LoadListing:
    requirements = {**context['hints'], **context['requirements']}
    if 'loadListing' in port_description:
        return LoadListing[port_description['loadListing']]
    elif 'outputBinding' in port_description and 'loadListing' in port_description['outputBinding']:
        return LoadListing[port_description['outputBinding']['loadListing']]
    elif 'LoadListingRequirement' in requirements and 'loadListing' in requirements['LoadListingRequirement']:
        return LoadListing[requirements['LoadListingRequirement']['loadListing']]
    else:
        return LoadListing.no_listing


def _get_name(name_prefix: str,
              cwl_name_prefix: str,
              element_id: str,
              preserve_cwl_prefix: bool = False) -> str:
    name = element_id.split('#')[-1]
    return (posixpath.join(posixpath.sep, name) if preserve_cwl_prefix else posixpath.join(
        name_prefix, posixpath.relpath(posixpath.join(posixpath.sep, name), cwl_name_prefix)))


def _get_path(element_id: str) -> str:
    path = element_id
    if '#' in path:
        path = path.split('#')[-1]
    if path.startswith('file://'):
        path = urllib.parse.unquote(path[7:])
    return path


def _get_schema_def_types(requirements: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
    return ({sd['name']: sd for sd in requirements['SchemaDefRequirement']['types']}
            if 'SchemaDefRequirement' in requirements else {})


def _get_secondary_files(cwl_element, default_required: bool) -> MutableSequence[SecondaryFile]:
    if not cwl_element:
        return []
    elif isinstance(cwl_element, MutableSequence):
        return [SecondaryFile(
            pattern=sf['pattern'],
            required=sf.get('required') if sf.get('required') is not None else default_required
        ) for sf in cwl_element]
    elif isinstance(cwl_element, MutableMapping):
        return [SecondaryFile(
            pattern=cwl_element['pattern'],
            required=cwl_element.get('required') if cwl_element.get('required') is not None else default_required)]


def _get_type_from_array(port_type: str):
    if isinstance(port_type, MutableMapping):
        if port_type['type'] == 'array':
            return _get_type_from_array(port_type['items'])
        else:
            return port_type['type']
    else:
        return port_type


def _percolate_port(port_name: str, *args) -> Port:
    for arg in args:
        if port_name in arg:
            port = arg[port_name]
            if isinstance(port, Port):
                return port
            else:
                return _percolate_port(port, *args)


def _process_docker_image(docker_requirement: MutableMapping[str, Any]) -> str:
    # Retrieve image
    if 'dockerPull' in docker_requirement:
        image_name = docker_requirement['dockerPull']
    elif 'dockerImageId' in docker_requirement:
        image_name = docker_requirement['dockerImageId']
    else:
        raise WorkflowDefinitionException(
            "DockerRequirements without `dockerPull` or `dockerImageId` are not supported yet")
    return image_name


def _process_docker_requirement(name: str,
                                target: Target,
                                context: MutableMapping[str, Any],
                                docker_requirement: MutableMapping[str, Any],
                                network_access: bool) -> Target:
    image_name = _process_docker_image(docker_requirement=docker_requirement)
    # Build configuration
    docker_config = {
        'image': image_name,
        'logDriver': 'none',
        'network': 'default' if network_access else 'none',
        'volume': ['{host}:{container}'.format(
            host=target.workdir,
            container=target.workdir)]}
    # Manage dockerOutputDirectory directive
    if 'dockerOutputDirectory' in docker_requirement:
        docker_config['workdir'] = docker_requirement['dockerOutputDirectory']
        context['output_directory'] = docker_config['workdir']
        local_dir = os.path.join(tempfile.gettempdir(), 'streamflow', random_name())
        os.makedirs(local_dir, exist_ok=True)
        docker_config['volume'].append('{host}:{container}'.format(
            host=local_dir,
            container=docker_config['workdir']))
    # Build step target
    deployment = DeploymentConfig(
        name=name,
        type='docker',
        config=docker_config)
    step_target = Target(
        deployment=deployment,
        service=image_name,
        workdir=target.workdir)
    return step_target


def _process_javascript_requirement(requirements: MutableMapping[str, Any]) -> (Optional[MutableSequence[Any]], bool):
    expression_lib = None
    full_js = False
    if 'InlineJavascriptRequirement' in requirements:
        full_js = True
        if 'expressionLib' in requirements['InlineJavascriptRequirement']:
            expression_lib = []
            for lib in requirements['InlineJavascriptRequirement']['expressionLib']:
                expression_lib.append(lib)
    return expression_lib, full_js


def _process_loop_transformers(step_name: str,
                               input_ports: MutableMapping[str, Port],
                               loop_input_ports: MutableMapping[str, Port],
                               transformers: MutableMapping[str, LoopValueFromTransformer],
                               input_dependencies: MutableMapping[str, Set[str]]):
    new_input_ports = {}
    for input_name, token_transformer in transformers.items():
        # Process inputs to attach ports
        for dep_name in input_dependencies[input_name]:
            if dep_name in input_ports:
                token_transformer.add_loop_input_port(
                    posixpath.relpath(dep_name, step_name), input_ports[dep_name])
        if input_name in loop_input_ports:
            token_transformer.add_loop_source_port(
                posixpath.relpath(input_name, step_name), loop_input_ports[input_name])
        # Put transformer output ports in input ports map
        new_input_ports[input_name] = token_transformer.get_output_port()
    return {**input_ports, **loop_input_ports, **new_input_ports}


def _process_transformers(step_name: str,
                          input_ports: MutableMapping[str, Port],
                          transformers: MutableMapping[str, Transformer],
                          input_dependencies: MutableMapping[str, Set[str]]):
    new_input_ports = {}
    for input_name, token_transformer in transformers.items():
        # Process inputs to attach ports
        for dep_name in input_dependencies[input_name]:
            if dep_name in input_ports:
                token_transformer.add_input_port(
                    posixpath.relpath(dep_name, step_name), input_ports[dep_name])
        # Put transformer output ports in input ports map
        new_input_ports[input_name] = token_transformer.get_output_port()
    return {**input_ports, **new_input_ports}


def _remap_path(path_processor,
                path: str,
                old_dir: str,
                new_dir: str) -> str:
    if ':/' in path:
        scheme = urllib.parse.urlsplit(path).scheme
        if scheme == 'file':
            return 'file://{}'.format(
                path_processor.join(new_dir, *os.path.relpath(
                    urllib.parse.unquote(path[7:]), old_dir).split(os.path.sep)))
        else:
            return path
    else:
        return path_processor.join(new_dir, *os.path.relpath(
            urllib.parse.unquote(path[7:]), old_dir).split(os.path.sep))


class CWLTranslator(object):

    def __init__(self,
                 context: StreamFlowContext,
                 name: str,
                 output_directory: str,
                 cwl_definition: cwltool.process.Process,
                 cwl_inputs: MutableMapping[str, Any],
                 workflow_config: WorkflowConfig,
                 loading_context: cwltool.context.LoadingContext):
        self.context: StreamFlowContext = context
        self.name: str = name
        self.output_directory: str = output_directory
        self.cwl_definition: cwltool.process.Process = cwl_definition
        self.cwl_inputs: MutableMapping[str, Any] = cwl_inputs
        self.default_map: MutableMapping[str, Any] = {}
        self.deployment_map: MutableMapping[str, DeployStep] = {}
        self.gather_map: MutableMapping[str, str] = {}
        self.input_ports: MutableMapping[str, Port] = {}
        self.loading_context: cwltool.context.LoadingContext = loading_context
        self.output_ports: MutableMapping[str, Union[str, Port]] = {}
        self.scatter: MutableMapping[str, Any] = {}
        self.workflow_config: WorkflowConfig = workflow_config

    def _get_deploy_step(self,
                         target: Target,
                         workflow: Workflow):
        if target.deployment.name not in self.deployment_map:
            self.deployment_map[target.deployment.name] = workflow.create_step(
                cls=DeployStep,
                name=posixpath.join("__deploy__", target.deployment.name),
                deployment_config=target.deployment)
        return self.deployment_map[target.deployment.name]

    def _get_input_port(self,
                        workflow: Workflow,
                        cwl_element: cwltool.workflow.Process,
                        element_input: MutableMapping[str, Any],
                        global_name: str,
                        port_name: str) -> Port:
        # Retrieve or create input port
        if global_name not in self.input_ports:
            self.input_ports[global_name] = workflow.create_port()
        input_port = self.input_ports[global_name]
        # If there is a default value, construct a default port block
        if 'default' in element_input:
            # Insert default port
            transformer_suffix = ('-wf-default-transformer' if isinstance(cwl_element, cwltool.workflow.Workflow)
                                  else '-cmd-default-transformer')
            input_port = self._handle_default_port(
                global_name=global_name,
                port_name=port_name,
                transformer_suffix=transformer_suffix,
                port=input_port,
                workflow=workflow,
                value=element_input['default'])
        # Return port
        return input_port

    def _get_source_port(self, workflow: Workflow, source_name: str) -> Port:
        if source_name not in self.input_ports:
            if source_name not in self.output_ports:
                self.output_ports[source_name] = workflow.create_port()
            return self.output_ports[source_name]
        else:
            return self.input_ports[source_name]

    def _get_target(self, name: str, target_type: str):
        path = PurePosixPath(name)
        target_config = self.workflow_config.propagate(path, target_type)
        workdir = target_config.get('workdir') if target_config is not None else None
        if target_config is not None:
            if 'deployment' in target_config:
                target_deployment = self.workflow_config.deplyoments[target_config['deployment']]
            else:
                target_deployment = self.workflow_config.deplyoments[target_config['model']]
                logger.warn("The `model` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                            "Use `deployment` instead.")
            locations = target_config.get('locations', None)
            if locations is None:
                locations = target_config.get('resources')
                if locations is not None:
                    logger.warn("The `resources` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                                "Use `locations` instead.")
                else:
                    locations = 1
            deployment = DeploymentConfig(
                name=target_deployment['name'],
                type=target_deployment['type'],
                config=target_deployment['config'],
                external=target_deployment.get('external', False),
                lazy=target_deployment.get('lazy', True),
                workdir=target_deployment.get('workdir'))
            target = Target(
                deployment=deployment,
                locations=locations,
                service=target_config.get('service'),
                workdir=workdir)
            return target
        else:
            return LocalTarget(workdir=workdir)

    def _handle_default_port(self,
                             global_name: str,
                             port_name: str,
                             transformer_suffix: str,
                             port: Optional[Port],
                             workflow: Workflow,
                             value: Any) -> Port:
        # Check output directory
        path = _get_path(self.cwl_definition.tool['id'])
        # Build default port
        default_port = workflow.create_port()
        self._inject_input(
            workflow=workflow,
            port_name="-".join([port_name, "default"]),
            global_name=global_name + transformer_suffix,
            port=default_port,
            output_directory=os.path.dirname(path),
            value=value)
        if port is not None:
            # Add default transformer
            transformer = workflow.create_step(
                cls=CWLDefaultTransformer,
                name=global_name + transformer_suffix,
                default_port=default_port)
            transformer.add_input_port(port_name, port)
            transformer.add_output_port(port_name, workflow.create_port())
            return transformer.get_output_port()
        else:
            return default_port

    def _inject_input(self,
                      workflow: Workflow,
                      global_name: str,
                      port_name,
                      port: Port,
                      output_directory: str,
                      value: Any) -> None:
        # Retrieve a local DeployStep
        target = self._get_target(global_name, 'port')
        deploy_step = self._get_deploy_step(target, workflow)
        # Remap path if target's workdir is defined
        connector = self.context.deployment_manager.get_connector(target.deployment.name)
        path_processor = utils.get_path_processor(connector)
        if isinstance(value, MutableMapping) and utils.get_token_class(value) in ['File', 'Directory']:
            target_config = self.workflow_config.propagate(PurePosixPath(global_name), 'port')
            if target_config:
                if 'location' in value:
                    value['location'] = _remap_path(
                        path_processor=path_processor,
                        path=value['location'],
                        old_dir=output_directory,
                        new_dir=target.workdir)
                if 'path' in value:
                    value['path'] = _remap_path(
                        path_processor=path_processor,
                        path=value['path'],
                        old_dir=output_directory,
                        new_dir=target.workdir)
        # Create a schedule step and connect it to the local DeployStep
        schedule_step = workflow.create_step(
            cls=ScheduleStep,
            name=posixpath.join(global_name + "-injector", "__schedule__"),
            connector_port=deploy_step.get_output_port(),
            input_directory=target.workdir or output_directory,
            output_directory=target.workdir or output_directory,
            tmp_directory=target.workdir or output_directory,
            target=target)
        # Create a CWLInputInjector step to process the input
        injector_step = workflow.create_step(
            cls=CWLInputInjectorStep,
            name=global_name + "-injector",
            job_port=schedule_step.get_output_port())
        # Create an input port and inject values
        input_port = workflow.create_port()
        input_port.put(Token(value=value))
        input_port.put(TerminationToken())
        # Connect input and output ports to the injector step
        injector_step.add_input_port(port_name, input_port)
        injector_step.add_output_port(port_name, port)

    def _inject_inputs(self,
                       workflow: Workflow):
        output_directory = None
        if self.cwl_inputs:
            # Compute output directory path
            path = _get_path(self.cwl_inputs['id'])
            output_directory = os.path.dirname(path)
        # Compute suffix
        default_suffix = ('-wf-default-transformer' if isinstance(self.cwl_definition, cwltool.workflow.Workflow)
                          else '-cmd-default-transformer')
        # Process externally provided inputs
        for global_name in self.input_ports:
            if (step := workflow.steps.get(
                    global_name + default_suffix, workflow.steps.get(global_name + "-token-transformer"))):
                step_name = posixpath.dirname(global_name)
                port_name = posixpath.relpath(global_name, step_name)
                input_port = step.get_input_port(port_name)
                # If an input is given for port, inject it
                if step_name == posixpath.sep and port_name in self.cwl_inputs:
                    self._inject_input(
                        workflow=workflow,
                        global_name=global_name,
                        port_name=port_name,
                        port=input_port,
                        output_directory=output_directory,
                        value=self.cwl_inputs[port_name])
        # Search empty unbound input ports
        for input_port in workflow.ports.values():
            if input_port.empty() and not input_port.get_input_steps():
                input_port.put(Token(value=None))
                input_port.put(TerminationToken())

    def _recursive_translate(self,
                             workflow: Workflow,
                             cwl_element: cwltool.process.Process,
                             context: MutableMapping[str, Any],
                             name_prefix: str,
                             cwl_name_prefix: str):
        # Update context
        current_context = copy.deepcopy(context)
        for hint in cwl_element.hints:
            current_context['hints'][hint['class']] = hint
        for requirement in cwl_element.requirements:
            current_context['requirements'][requirement['class']] = requirement
        # In the root process, override requirements when provided in the input file
        if name_prefix == posixpath.sep:
            req_string = 'https://w3id.org/cwl/cwl#requirements'
            if req_string in self.cwl_inputs:
                current_context['requirements'] = {req['class']: req for req in self.cwl_inputs[req_string]}
        # Dispatch element
        if isinstance(cwl_element, cwltool.workflow.Workflow):
            self._translate_workflow(
                workflow=workflow,
                cwl_element=cwl_element,
                context=current_context,
                name_prefix=name_prefix,
                cwl_name_prefix=cwl_name_prefix)
        elif isinstance(cwl_element, cwltool.workflow.WorkflowStep):
            self._translate_workflow_step(
                workflow=workflow,
                cwl_element=cwl_element,
                context=current_context,
                name_prefix=name_prefix,
                cwl_name_prefix=cwl_name_prefix)
        elif isinstance(cwl_element, cwltool.command_line_tool.CommandLineTool):
            self._translate_command_line_tool(
                workflow=workflow,
                cwl_element=cwl_element,
                context=current_context,
                name_prefix=name_prefix,
                cwl_name_prefix=cwl_name_prefix)
        elif isinstance(cwl_element, cwltool.command_line_tool.ExpressionTool):
            self._translate_command_line_tool(
                workflow=workflow,
                cwl_element=cwl_element,
                context=current_context,
                name_prefix=name_prefix,
                cwl_name_prefix=cwl_name_prefix)
        else:
            raise WorkflowDefinitionException(
                "Definition of type " + type(cwl_element).__class__.__name__ + " not supported")

    def _translate_command_line_tool(self,
                                     workflow: Workflow,
                                     cwl_element: Union[cwltool.command_line_tool.CommandLineTool,
                                                        cwltool.command_line_tool.ExpressionTool],
                                     context: MutableMapping[str, Any],
                                     name_prefix: str,
                                     cwl_name_prefix: str):
        logger.debug("Translating {} {}".format(cwl_element.tool.__class__.__name__, name_prefix))
        # Extract custom types if present
        requirements = {**context['hints'], **context['requirements']}
        schema_def_types = _get_schema_def_types(requirements)
        # Process InlineJavascriptRequirement
        expression_lib, full_js = _process_javascript_requirement(requirements)
        # Retrieve target
        target = self._get_target(name_prefix, 'step')
        # Process DockerRequirement
        if 'DockerRequirement' in requirements:
            network_access = (requirements['NetworkAccess']['networkAccess'] if 'NetworkAccess' in requirements
                              else False)
            if target.deployment.name == LOCAL_LOCATION:
                target = _process_docker_requirement(
                    name=posixpath.join(name_prefix, 'docker-requirement'),
                    target=target,
                    context=context,
                    docker_requirement=requirements['DockerRequirement'],
                    network_access=network_access)
            elif 'image' in target.deployment.config and target.deployment.config['image'] == '':
                # Overwite image configuration
                image_name = _process_docker_image(docker_requirement=requirements['DockerRequirement'])
                target.deployment.config['image'] = image_name
        # Create DeployStep to initialise the execution environment
        deploy_step = self._get_deploy_step(target, workflow)
        # Create a schedule step and connect it to the DeployStep
        schedule_step = workflow.create_step(
            cls=ScheduleStep,
            name=posixpath.join(name_prefix, '__schedule__'),
            connector_port=deploy_step.get_output_port(),
            target=target,
            hardware_requirement=_get_hardware_requirement(requirements, expression_lib, full_js),
            output_directory=context.get('output_directory'))
        # Create the ExecuteStep and connect it to the ScheduleStep
        step = workflow.create_step(
            cls=ExecuteStep,
            name=name_prefix,
            job_port=schedule_step.get_output_port())
        # Process inputs
        input_ports = {}
        token_transformers = []
        for element_input in cwl_element.tool['inputs']:
            global_name = _get_name(
                name_prefix, cwl_name_prefix, element_input['id'])
            port_name = posixpath.relpath(global_name, name_prefix)
            # Retrieve or create input port
            input_port = self._get_input_port(
                workflow=workflow,
                cwl_element=cwl_element,
                element_input=element_input,
                global_name=global_name,
                port_name=port_name)
            # Add a token transformer step to process inputs
            token_transformer = _create_token_transformer(
                name=global_name + "-token-transformer",
                port_name=port_name,
                workflow=workflow,
                cwl_element=element_input,
                cwl_name_prefix=posixpath.join(cwl_name_prefix, port_name),
                schema_def_types=schema_def_types,
                format_graph=self.loading_context.loader.graph,
                context=context,
                only_propagate_secondary_files=(name_prefix != '/'))
            # Add the output port as an input of the schedule step
            schedule_step.add_input_port(port_name, token_transformer.get_output_port())
            # Create a TransferStep
            transfer_step = workflow.create_step(
                cls=CWLTransferStep,
                name=posixpath.join(name_prefix, "__transfer__", port_name),
                job_port=schedule_step.get_output_port())
            transfer_step.add_input_port(port_name, token_transformer.get_output_port())
            transfer_step.add_output_port(port_name, workflow.create_port())
            # Connect the transfer step with the ExecuteStep
            step.add_input_port(port_name, transfer_step.get_output_port(port_name))
            # Store input port and token transformer
            input_ports[port_name] = input_port
            token_transformers.append(token_transformer)
        # Add input ports to token transformers
        for port_name, input_port in input_ports.items():
            for token_transformer in token_transformers:
                token_transformer.add_input_port(port_name, input_port)
        # Process outputs
        for element_output in cwl_element.tool['outputs']:
            global_name = _get_name(
                name_prefix, cwl_name_prefix, element_output['id'])
            port_name = posixpath.relpath(global_name, name_prefix)
            # Retrieve or create output port
            if global_name not in self.output_ports:
                self.output_ports[global_name] = workflow.create_port()
            output_port = self.output_ports[global_name]
            # If the port is bound to a remote target, add the connector dependency
            if self.workflow_config.propagate(PurePosixPath(global_name), 'port'):
                port_target = self._get_target(global_name, 'port')
                output_deploy_step = self._get_deploy_step(port_target, workflow)
                step.add_input_port(port_name + '__connector__', output_deploy_step.get_output_port())
                step.output_connectors[port_name] = port_name + '__connector__'
            else:
                port_target = None
            # Add output port to ExecuteStep
            step.add_output_port(
                name=port_name,
                port=output_port,
                output_processor=_create_command_output_processor(
                    port_name=port_name,
                    workflow=workflow,
                    port_target=port_target,
                    port_type=element_output['type'],
                    cwl_element=element_output,
                    cwl_name_prefix=posixpath.join(cwl_name_prefix, port_name),
                    schema_def_types=schema_def_types,
                    context=context))
        if isinstance(cwl_element, cwltool.command_line_tool.CommandLineTool):
            # Process command
            step.command = _create_command(
                cwl_element=cwl_element,
                cwl_name_prefix=cwl_name_prefix,
                schema_def_types=schema_def_types,
                context=context,
                step=step)
            # Process ToolTimeLimit
            if 'ToolTimeLimit' in requirements:
                step.command.time_limit = requirements['ToolTimeLimit']['timelimit']
                if step.command.time_limit < 0:
                    raise WorkflowDefinitionException('Invalid time limit for step {step}'.format(step=name_prefix))
        elif isinstance(cwl_element, cwltool.command_line_tool.ExpressionTool):
            if 'expression' in cwl_element.tool:
                step.command = CWLExpressionCommand(step, cwl_element.tool['expression'])
        # Add JS requirements
        step.command.expression_lib = expression_lib
        step.command.full_js = full_js
        # Process streams
        if 'stdin' in cwl_element.tool:
            step.command.stdin = cwl_element.tool['stdin']
        if 'stdout' in cwl_element.tool:
            step.command.stdout = cwl_element.tool['stdout']
        if 'stderr' in cwl_element.tool:
            step.command.stderr = cwl_element.tool['stderr']

    def _translate_workflow(self,
                            workflow: Workflow,
                            cwl_element: cwltool.workflow.Workflow,
                            context: MutableMapping[str, Any],
                            name_prefix: str,
                            cwl_name_prefix: str):
        step_name = name_prefix
        logger.debug("Translating Workflow {}".format(step_name))
        # Extract custom types if present
        requirements = {**context['hints'], **context['requirements']}
        schema_def_types = _get_schema_def_types(requirements)
        # Extract JavaScript requirements
        expression_lib, full_js = _process_javascript_requirement(requirements)
        # Process inputs to create steps
        input_ports = {}
        token_transformers = {}
        input_dependencies = {}
        for element_input in cwl_element.tool['inputs']:
            global_name = _get_name(step_name, cwl_name_prefix, element_input['id'])
            port_name = posixpath.relpath(global_name, step_name)
            # Retrieve or create input port
            input_ports[global_name] = self._get_input_port(
                workflow=workflow,
                cwl_element=cwl_element,
                element_input=element_input,
                global_name=global_name,
                port_name=port_name)
            # Create token transformer step
            token_transformers[global_name] = _create_token_transformer(
                name=global_name + "-token-transformer",
                port_name=port_name,
                workflow=workflow,
                cwl_element=element_input,
                cwl_name_prefix=posixpath.join(cwl_name_prefix, port_name),
                schema_def_types=schema_def_types,
                format_graph=self.loading_context.loader.graph,
                context=context,
                only_propagate_secondary_files=(name_prefix != '/'))
            # Process dependencies
            local_deps = resolve_dependencies(
                expression=element_input.get('format'),
                full_js=full_js,
                expression_lib=expression_lib)
            if 'secondaryFiles' in element_input:
                for secondary_file in element_input['secondaryFiles']:
                    local_deps.update(
                        resolve_dependencies(
                            expression=secondary_file.get('pattern'),
                            full_js=full_js,
                            expression_lib=expression_lib),
                        resolve_dependencies(
                            expression=secondary_file.get('required'),
                            full_js=full_js,
                            expression_lib=expression_lib))
            input_dependencies[global_name] = set.union(
                {global_name},
                {posixpath.join(step_name, d) for d in local_deps})
        # Process inputs again to attach ports
        input_ports = _process_transformers(
            step_name=step_name,
            input_ports=input_ports,
            transformers=token_transformers,
            input_dependencies=input_dependencies)
        # Save input ports in the global map
        for input_name in token_transformers:
            self.input_ports[input_name] = input_ports[input_name]
        # Process outputs
        for element_output in cwl_element.tool['outputs']:
            global_name = _get_name(name_prefix, cwl_name_prefix, element_output['id'])
            # If outputSource element is a list, the output element can depend on multiple ports
            if isinstance(element_output['outputSource'], MutableSequence):
                # If the list contains only one element and no `linkMerge` or `pickValue` are specified
                if (len(element_output['outputSource']) == 1 and
                        'linkMerge' not in element_output and
                        'pickValue' not in element_output):
                    # Treat it as a singleton
                    source_name = _get_name(
                        name_prefix, cwl_name_prefix, element_output['outputSource'][0])
                    # If the output source is an input port, link the output to the input
                    if source_name in self.input_ports:
                        self.output_ports[global_name] = self.input_ports[source_name]
                    # Otherwise, simply propagate the output port
                    else:
                        self.output_ports[source_name] = self._get_source_port(workflow, global_name)
                # Otherwise, create a ListMergeCombinator
                else:
                    source_names = [_get_name(name_prefix, cwl_name_prefix, src)
                                    for src in element_output['outputSource']]
                    ports = {n: self._get_source_port(workflow, n) for n in source_names}
                    _create_list_merger(
                        name=global_name,
                        workflow=workflow,
                        ports=ports,
                        output_port=self._get_source_port(workflow, global_name),
                        link_merge=element_output.get('linkMerge'),
                        pick_value=element_output.get('pickValue'))
            # Otherwise, the output element depends on a single output port
            else:
                source_name = _get_name(
                    name_prefix, cwl_name_prefix, element_output['outputSource'])
                # If `pickValue` is specified, create a ListMergeCombinator
                if 'pickValue' in element_output:
                    source_port = self._get_source_port(workflow, source_name)
                    _create_list_merger(
                        name=global_name,
                        workflow=workflow,
                        ports={source_name: source_port},
                        output_port=self._get_source_port(workflow, global_name),
                        link_merge=element_output.get('linkMerge'),
                        pick_value=element_output.get('pickValue'))
                else:
                    # If the output source is an input port, link the output to the input
                    if source_name in self.input_ports:
                        self.output_ports[global_name] = self.input_ports[source_name]
                    # Otherwise, simply propagate the output port
                    else:
                        if global_name not in self.output_ports:
                            self.output_ports[global_name] = workflow.create_port()
                        self.output_ports[source_name] = self.output_ports[global_name]
        # Process steps
        for step in cwl_element.steps:
            self._recursive_translate(
                workflow=workflow,
                cwl_element=step,
                context=context,
                name_prefix=name_prefix,
                cwl_name_prefix=cwl_name_prefix)

    def _translate_workflow_step(self,
                                 workflow: Workflow,
                                 cwl_element: cwltool.workflow.WorkflowStep,
                                 context: MutableMapping[str, Any],
                                 name_prefix: str,
                                 cwl_name_prefix: str):
        # Process content
        step_name = _get_name(name_prefix, cwl_name_prefix, cwl_element.id)
        logger.debug("Translating WorkflowStep {}".format(step_name))
        cwl_step_name = _get_name(
            name_prefix, cwl_name_prefix, cwl_element.id, preserve_cwl_prefix=True)
        # Extract requirements
        for hint in cwl_element.embedded_tool.hints:
            context['hints'][hint['class']] = hint
        for requirement in cwl_element.embedded_tool.requirements:
            context['requirements'][requirement['class']] = requirement
        requirements = {**context['hints'], **context['requirements']}
        # Extract JavaScript requirements
        expression_lib, full_js = _process_javascript_requirement(requirements)
        # Find scatter elements
        if isinstance(cwl_element.tool.get('scatter'), str):
            scatter_inputs = [_get_name(
                step_name, cwl_step_name, cwl_element.tool.get('scatter'))]
        else:
            scatter_inputs = [_get_name(step_name, cwl_step_name, n)
                              for n in cwl_element.tool.get('scatter', [])]
        # Process inputs
        input_ports = {}
        value_from_transformers = {}
        input_dependencies = {}
        for element_input in cwl_element.tool['inputs']:
            self._translate_workflow_step_input(
                workflow=workflow,
                context=context,
                element_id=cwl_element.id,
                element_input=element_input,
                name_prefix=name_prefix,
                cwl_name_prefix=cwl_name_prefix,
                scatter_inputs=scatter_inputs,
                requirements=requirements,
                input_ports=input_ports,
                value_from_transformers=value_from_transformers,
                input_dependencies=input_dependencies)
        # Process loop inputs
        element_requirements = {
            **{h['class']: h for h in cwl_element.embedded_tool.hints},
            **{r['class']: r for r in cwl_element.embedded_tool.requirements}}
        if 'http://commonwl.org/cwltool#Loop' in element_requirements:
            loop_requirement = element_requirements['http://commonwl.org/cwltool#Loop']
            # Validate loop requirement
            if 'when' in cwl_element.tool:
                raise WorkflowDefinitionException(
                    'cwltool:Loop clause is not compatible with the `when` directive')
            if scatter_inputs:
                raise WorkflowDefinitionException(
                    'cwltool:Loop requirement is not compatible with the `scatter` directive')
            if 'loopWhen' not in loop_requirement:
                raise WorkflowDefinitionException(
                    'The `loopWhen` is required for cwltool:Loop requirement')
            # Build combinator
            loop_combinator = LoopCombinator(
                workflow=workflow,
                name=step_name + "-loop-combinator")
            for global_name in input_ports:
                # Decouple loop ports through a forwarder
                port_name = posixpath.relpath(global_name, step_name)
                loop_forwarder = workflow.create_step(
                    cls=ForwardTransformer,
                    name=global_name + "-input-forward-transformer")
                loop_forwarder.add_input_port(port_name, input_ports[global_name])
                input_ports[global_name] = workflow.create_port()
                loop_forwarder.add_output_port(port_name, input_ports[global_name])
                # Add item to combinator
                loop_combinator.add_item(posixpath.relpath(global_name, step_name))
            # Create a combinator step and add all inputs to it
            combinator_step = workflow.create_step(
                cls=LoopCombinatorStep,
                name=step_name + "-loop-combinator",
                combinator=loop_combinator)
            for global_name in input_ports:
                port_name = posixpath.relpath(global_name, step_name)
                combinator_step.add_input_port(port_name, input_ports[global_name])
                combinator_step.add_output_port(port_name, workflow.create_port())
            # Create loop conditional step
            loop_conditional_step = workflow.create_step(
                cls=CWLLoopConditionalStep,
                name=step_name + "-loop-when",
                expression=loop_requirement['loopWhen'],
                expression_lib=expression_lib,
                full_js=full_js)
            # Add inputs to conditional step
            for global_name in input_ports:
                port_name = posixpath.relpath(global_name, step_name)
                loop_conditional_step.add_input_port(port_name, combinator_step.get_output_port(port_name))
                input_ports[global_name] = workflow.create_port()
                loop_conditional_step.add_output_port(port_name, input_ports[global_name])
        # Retrieve scatter method (default to dotproduct)
        scatter_method = cwl_element.tool.get('scatterMethod', 'dotproduct')
        # If there are scatter inputs
        if scatter_inputs:
            # If any scatter input is null, propagate an empty array on the output ports
            empty_scatter_conditional_step = workflow.create_step(
                cls=CWLEmptyScatterConditionalStep,
                name=step_name + "-empty-scatter-condition",
                scatter_method=scatter_method)
            for global_name in scatter_inputs:
                port_name = posixpath.relpath(global_name, step_name)
                empty_scatter_conditional_step.add_input_port(port_name, input_ports[global_name])
                input_ports[global_name] = workflow.create_port()
                empty_scatter_conditional_step.add_output_port(port_name, input_ports[global_name])
            # If there are multiple scatter inputs, configure combinator
            scatter_combinator = None
            if len(scatter_inputs) > 1:
                # Build combinator
                if scatter_method == 'dotproduct':
                    scatter_combinator = DotProductCombinator(
                        workflow=workflow,
                        name=step_name + "-scatter-combinator")
                    for global_name in scatter_inputs:
                        scatter_combinator.add_item(posixpath.relpath(global_name, step_name))
                else:
                    scatter_combinator = CartesianProductCombinator(
                        workflow=workflow,
                        name=step_name + "-scatter-combinator")
                    for global_name in scatter_inputs:
                        scatter_combinator.add_item(posixpath.relpath(global_name, step_name))
            # If there are both scatter and non-scatter inputs
            if len(scatter_inputs) < len(input_ports):
                scatter_combinator = _create_residual_combinator(
                    workflow=workflow,
                    step_name=step_name,
                    inner_combinator=scatter_combinator,
                    inner_inputs=scatter_inputs,
                    input_ports=input_ports)
            # If there are scatter inputs, process them
            for global_name in scatter_inputs:
                port_name = posixpath.relpath(global_name, step_name)
                scatter_step = workflow.create_step(
                    cls=ScatterStep,
                    name=global_name + "-scatter")
                scatter_step.add_input_port(port_name, input_ports[global_name])
                input_ports[global_name] = workflow.create_port()
                scatter_step.add_output_port(port_name, input_ports[global_name])
            # If there is a scatter combinator, create a combinator step and add all inputs to it
            if scatter_combinator:
                combinator_step = workflow.create_step(
                    cls=CombinatorStep,
                    name=step_name + "-scatter-combinator",
                    combinator=scatter_combinator)
                for global_name in input_ports:
                    port_name = posixpath.relpath(global_name, step_name)
                    combinator_step.add_input_port(port_name, input_ports[global_name])
                    input_ports[global_name] = workflow.create_port()
                    combinator_step.add_output_port(port_name, input_ports[global_name])
        # Process inputs again to attach ports to transformers
        input_ports = _process_transformers(
            step_name=step_name,
            input_ports=input_ports,
            transformers=value_from_transformers,
            input_dependencies=input_dependencies)
        # Save input ports in the global map
        self.input_ports = {**self.input_ports, **input_ports}
        # Process condition
        conditional_step = None
        if 'when' in cwl_element.tool:
            # Create conditional step
            conditional_step = workflow.create_step(
                cls=CWLConditionalStep,
                name=step_name + "-when",
                expression=cwl_element.tool['when'],
                expression_lib=expression_lib,
                full_js=full_js)
            # Add inputs and outputs to conditional step
            for global_name in input_ports:
                port_name = posixpath.relpath(global_name, step_name)
                conditional_step.add_input_port(port_name, self.input_ports[global_name])
                self.input_ports[global_name] = workflow.create_port()
                conditional_step.add_output_port(port_name, self.input_ports[global_name])
        # Process outputs
        external_output_ports = {}
        internal_output_ports = {}
        for element_output in cwl_element.tool['outputs']:
            global_name = _get_name(
                step_name, cwl_step_name, element_output['id'])
            port_name = posixpath.relpath(global_name, step_name)
            # Retrieve or create output port
            if global_name not in self.output_ports:
                self.output_ports[global_name] = workflow.create_port()
            external_output_ports[global_name] = self.output_ports[global_name]
            internal_output_ports[global_name] = self.output_ports[global_name]
            # If there are scatter inputs
            if scatter_inputs:
                # Perform a gather on the outputs
                if scatter_method == 'nested_crossproduct':
                    gather_steps = []
                    internal_output_ports[global_name] = workflow.create_port()
                    gather_input_port = internal_output_ports[global_name]
                    for scatter_input in scatter_inputs:
                        scatter_port_name = posixpath.relpath(scatter_input, step_name)
                        gather_steps.append(workflow.create_step(
                            cls=GatherStep,
                            name=global_name + "-gather-" + scatter_port_name))
                        gather_steps[-1].add_input_port(port_name, gather_input_port)
                        gather_steps[-1].add_output_port(
                            name=port_name,
                            port=(external_output_ports[global_name] if len(gather_steps) == len(scatter_inputs)
                                  else workflow.create_port()))
                        gather_input_port = gather_steps[-1].get_output_port()
                else:
                    gather_step = workflow.create_step(
                        cls=GatherStep,
                        name=global_name + "-gather",
                        depth=1 if scatter_method == 'dotproduct' else len(scatter_inputs))
                    internal_output_ports[global_name] = workflow.create_port()
                    gather_step.add_input_port(port_name, internal_output_ports[global_name])
                    gather_step.add_output_port(port_name, external_output_ports[global_name])
                # Add the output port as a skip port in the empty scatter conditional step
                empty_scatter_conditional_step = cast(
                    CWLConditionalStep, workflow.steps[step_name + "-empty-scatter-condition"])
                empty_scatter_conditional_step.add_skip_port(port_name, external_output_ports[global_name])
            # Add skip ports if there is a condition
            if 'when' in cwl_element.tool:
                cast(CWLConditionalStep, conditional_step).add_skip_port(
                    port_name, internal_output_ports[global_name])
        # Process loop outputs
        if 'http://commonwl.org/cwltool#Loop' in element_requirements:
            loop_requirement = element_requirements['http://commonwl.org/cwltool#Loop']
            output_method = loop_requirement.get('outputMethod', 'last')
            # Retreive loop steps
            loop_conditional_step = cast(CWLLoopConditionalStep, workflow.steps[step_name + "-loop-when"])
            combinator_step = workflow.steps[step_name + "-loop-combinator"]
            # Create a loop termination combinator
            loop_terminator_combinator = LoopTerminationCombinator(
                workflow=workflow,
                name=step_name + "-loop-termination-combinator")
            loop_terminator_step = workflow.create_step(
                cls=CombinatorStep,
                name=step_name + "-loop-terminator",
                combinator=loop_terminator_combinator)
            for port_name, port in combinator_step.get_input_ports().items():
                loop_terminator_step.add_output_port(port_name, port)
                loop_terminator_combinator.add_output_item(port_name)
            # Add outputs to conditional step
            for global_name in internal_output_ports:
                port_name = posixpath.relpath(global_name, step_name)
                # Create loop forwarder
                loop_forwarder = workflow.create_step(
                    cls=ForwardTransformer,
                    name=global_name + "-output-forward-transformer")
                internal_output_ports[global_name] = workflow.create_port()
                loop_forwarder.add_input_port(port_name, internal_output_ports[global_name])
                self.output_ports[global_name] = workflow.create_port()
                loop_forwarder.add_output_port(port_name, self.output_ports[global_name])
                # Create loop output step
                loop_output_step = workflow.create_step(
                    cls=CWLLoopOutputLastStep if output_method == 'last' else CWLLoopOutputAllStep,
                    name=global_name + "-loop-output")
                loop_output_step.add_input_port(port_name, loop_forwarder.get_output_port())
                loop_conditional_step.add_skip_port(port_name, loop_forwarder.get_output_port())
                loop_output_step.add_output_port(port_name, external_output_ports[global_name])
                loop_terminator_step.add_input_port(port_name, external_output_ports[global_name])
                loop_terminator_combinator.add_item(port_name)
            # Process inputs
            loop_input_ports = {}
            loop_value_from_transformers = {}
            loop_input_dependencies = {}
            for loop_input in loop_requirement.get('loop', []):
                # Pre-process the `loopSource` field to avoid inconsistencies
                loop_input['source'] = loop_input.get('loopSource', loop_input['id'])
                self._translate_workflow_step_input(
                    workflow=workflow,
                    context=context,
                    element_id=cwl_element.id,
                    element_input=loop_input,
                    name_prefix=name_prefix,
                    cwl_name_prefix=cwl_name_prefix,
                    scatter_inputs=scatter_inputs,
                    requirements=requirements,
                    input_ports=loop_input_ports,
                    value_from_transformers=loop_value_from_transformers,
                    input_dependencies=loop_input_dependencies,
                    inner_steps_prefix='-loop',
                    value_from_transformer_cls=LoopValueFromTransformer)
            # Process inputs again to attach ports to transformers
            loop_input_ports = _process_loop_transformers(
                step_name=step_name,
                input_ports=input_ports,
                loop_input_ports=loop_input_ports,
                transformers=loop_value_from_transformers,
                input_dependencies=loop_input_dependencies)
            # Connect loop outputs to loop inputs
            for global_name in input_ports:
                # Create loop output step
                port_name = posixpath.relpath(global_name, step_name)
                loop_forwarder = workflow.create_step(
                    cls=ForwardTransformer,
                    name=global_name + "-back-propagation-transformer")
                loop_forwarder.add_input_port(
                    port_name, loop_input_ports.get(global_name, loop_conditional_step.get_output_port(port_name)))
                loop_forwarder.add_output_port(port_name, combinator_step.get_input_port(port_name))
        # Add skip ports if there is a condition
        if 'when' in cwl_element.tool:
            for element_output in cwl_element.tool['outputs']:
                global_name = _get_name(step_name, cwl_step_name, element_output['id'])
                port_name = posixpath.relpath(global_name, step_name)
                skip_port = (external_output_ports[global_name]
                             if 'http://commonwl.org/cwltool#Loop' in element_requirements
                             else internal_output_ports[global_name])
                cast(CWLConditionalStep, conditional_step).add_skip_port(port_name, skip_port)
        # Update output ports with the internal ones
        self.output_ports = {**self.output_ports, **internal_output_ports}
        # Process inner element
        run_command = cwl_element.tool['run']
        if isinstance(run_command, MutableMapping):
            if ':' in cwl_element.embedded_tool.tool['id'].split('#')[-1]:
                inner_cwl_name_prefix = posixpath.join(cwl_step_name, 'run')
            else:
                inner_cwl_name_prefix = _get_name(
                    name_prefix, cwl_name_prefix, cwl_element.embedded_tool.tool['id'], preserve_cwl_prefix=True)
        else:
            inner_cwl_name_prefix = (_get_name(posixpath.sep, posixpath.sep, cwl_element.embedded_tool.tool['id'])
                                     if '#' in cwl_element.embedded_tool.tool['id'] else posixpath.sep)
        self._recursive_translate(
            workflow=workflow,
            cwl_element=cwl_element.embedded_tool,
            context=context,
            name_prefix=step_name,
            cwl_name_prefix=inner_cwl_name_prefix)
        # Update output ports with the external ones
        self.output_ports = {**self.output_ports, **external_output_ports}

    def _translate_workflow_step_input(self,
                                       workflow: Workflow,
                                       context: MutableMapping[str, Any],
                                       element_id: str,
                                       element_input: MutableMapping[str, Any],
                                       name_prefix: str,
                                       cwl_name_prefix: str,
                                       scatter_inputs: MutableSequence[str],
                                       requirements: MutableMapping[str, Any],
                                       input_ports: MutableMapping[str, Port],
                                       value_from_transformers: MutableMapping[str, ValueFromTransformer],
                                       input_dependencies: MutableMapping[str, Any],
                                       inner_steps_prefix: str = '',
                                       value_from_transformer_cls: Type[ValueFromTransformer] = ValueFromTransformer):
        # Extract custom types if present
        schema_def_types = _get_schema_def_types(requirements)
        # Extract JavaScript requirements
        expression_lib, full_js = _process_javascript_requirement(requirements)
        # Extract names
        step_name = _get_name(name_prefix, cwl_name_prefix, element_id)
        cwl_step_name = _get_name(
            name_prefix, cwl_name_prefix, element_id, preserve_cwl_prefix=True)
        global_name = _get_name(
            step_name, cwl_step_name, element_input['id'])
        port_name = posixpath.relpath(global_name, step_name)
        # Adjust type to handle scatter
        if global_name in scatter_inputs:
            element_input = {**element_input, **{'type': element_input['type']['items']}}
        # If element contains `valueFrom` directive
        if 'valueFrom' in element_input:
            # Create a ValueFromTransformer
            value_from_transformers[global_name] = workflow.create_step(
                cls=value_from_transformer_cls,
                name=global_name + inner_steps_prefix + "-value-from-transformer",
                processor=_create_token_processor(
                    port_name=port_name,
                    workflow=workflow,
                    port_type=element_input.get('type', 'Any'),
                    cwl_element=element_input,
                    cwl_name_prefix=posixpath.join(cwl_step_name, port_name),
                    schema_def_types=schema_def_types,
                    format_graph=self.loading_context.loader.graph,
                    context=context,
                    check_type=False),
                port_name=port_name,
                expression_lib=expression_lib,
                full_js=full_js,
                value_from=element_input['valueFrom'])
            value_from_transformers[global_name].add_output_port(port_name, workflow.create_port())
            # Retrieve dependencies
            local_deps = resolve_dependencies(
                expression=element_input.get('valueFrom'),
                full_js=full_js,
                expression_lib=expression_lib)
            input_dependencies[global_name] = set.union(
                {global_name} if 'source' in element_input or 'default' in element_input else set(),
                {posixpath.join(step_name, d) for d in local_deps}) or {global_name}
        # If `source` entry is present, process output dependencies
        if 'source' in element_input:
            # If source element is a list, the input element can depend on multiple ports
            if isinstance(element_input['source'], MutableSequence):
                # If the list contains only one element and no `linkMerge` is specified, treat it as a singleton
                if (len(element_input['source']) == 1 and
                        'linkMerge' not in element_input and
                        'pickValue' not in element_input):
                    source_name = _get_name(
                        name_prefix, cwl_name_prefix, element_input['source'][0])
                    source_port = self._get_source_port(workflow, source_name)
                    # If there is a default value, construct a default port block
                    if 'default' in element_input:
                        # Insert default port
                        source_port = self._handle_default_port(
                            global_name=global_name,
                            port_name=port_name,
                            transformer_suffix=inner_steps_prefix + "-step-default-transformer",
                            port=source_port,
                            workflow=workflow,
                            value=element_input['default'])
                    # Add source port to the list of input ports for the current step
                    input_ports[global_name] = source_port
                # Otherwise, create a ListMergeCombinator
                else:
                    source_names = [_get_name(name_prefix, cwl_name_prefix, src)
                                    for src in element_input['source']]
                    ports = {n: self._get_source_port(workflow, n) for n in source_names}
                    list_merger = _create_list_merger(
                        name=global_name + inner_steps_prefix + "-list-merge-combinator",
                        workflow=workflow,
                        ports=ports,
                        link_merge=element_input.get('linkMerge'),
                        pick_value=element_input.get('pickValue'))
                    # Add ListMergeCombinator output port to the list of input ports for the current step
                    input_ports[global_name] = list_merger.get_output_port()
            # Otherwise, the input element depends on a single output port
            else:
                source_name = _get_name(
                    name_prefix, cwl_name_prefix, element_input['source'])
                source_port = self._get_source_port(workflow, source_name)
                # If there is a default value, construct a default port block
                if 'default' in element_input:
                    # Insert default port
                    source_port = self._handle_default_port(
                        global_name=global_name,
                        port_name=port_name,
                        transformer_suffix=inner_steps_prefix + "-step-default-transformer",
                        port=source_port,
                        workflow=workflow,
                        value=element_input['default'])
                # Add source port to the list of input ports for the current step
                input_ports[global_name] = source_port
        # Otherwise, search for default values
        elif 'default' in element_input:
            input_ports[global_name] = self._handle_default_port(
                global_name=global_name,
                port_name=port_name,
                transformer_suffix=inner_steps_prefix + "-step-default-transformer",
                port=None,
                workflow=workflow,
                value=element_input['default'])
        # Otherwise, inject a synthetic port into the workflow
        else:
            input_ports[global_name] = workflow.create_port()

    def translate(self) -> Workflow:
        workflow = Workflow(
            context=self.context,
            type='cwl',
            name=self.name)
        # Create context
        context = _create_context()
        # Compute root prefix
        workflow_id = self.cwl_definition.tool['id']
        cwl_root_prefix = (_get_name(posixpath.sep, posixpath.sep, workflow_id) if '#' in workflow_id
                           else posixpath.sep)
        # Register data locations for config files
        path = _get_path(self.cwl_definition.tool['id'])
        self.context.data_manager.register_path(
            deployment=LOCAL_LOCATION,
            location=LOCAL_LOCATION,
            path=path,
            relpath=os.path.basename(path))
        if self.cwl_inputs:
            path = _get_path(self.cwl_inputs['id'])
            self.context.data_manager.register_path(
                deployment=LOCAL_LOCATION,
                location=LOCAL_LOCATION,
                path=path,
                relpath=os.path.basename(path))
        # Build workflow graph
        self._recursive_translate(
            workflow=workflow,
            cwl_element=self.cwl_definition,
            context=context,
            name_prefix=posixpath.sep,
            cwl_name_prefix=cwl_root_prefix)
        # Inject initial inputs
        self._inject_inputs(workflow)
        # Extract requirements
        for hint in self.cwl_definition.hints:
            context['hints'][hint['class']] = hint
        for requirement in self.cwl_definition.requirements:
            context['requirements'][requirement['class']] = requirement
        requirements = {**context['hints'], **context['requirements']}
        # Extract workflow outputs
        cwl_elements = {_get_name('/', cwl_root_prefix, element['id']): element
                        for element in self.cwl_definition.tool.get('outputs', [])}
        for output_name, output_value in self.output_ports.items():
            if output_name.lstrip(posixpath.sep).count(posixpath.sep) == 0:
                if port := _percolate_port(output_name, self.output_ports):
                    port_name = output_name.lstrip(posixpath.sep)
                    # Retrieve a local DeployStep
                    target = LocalTarget()
                    deploy_step = self._get_deploy_step(target, workflow)
                    # Create a transformer to enforce deep listing in folders
                    expression_lib, full_js = _process_javascript_requirement(requirements)
                    # Search for dependencies in format expression
                    format_deps = resolve_dependencies(
                        expression=cwl_elements[output_name].get('format'),
                        full_js=full_js,
                        expression_lib=expression_lib)
                    # Build transformer step
                    transformer_step = workflow.create_step(
                        cls=CWLTokenTransformer,
                        name=posixpath.join(port_name + "-collector-transformer"),
                        port_name=port_name,
                        processor=_create_token_processor(
                            port_name=port_name,
                            workflow=workflow,
                            port_type=cwl_elements[output_name]['type'],
                            cwl_element=cwl_elements[output_name],
                            cwl_name_prefix=posixpath.join(cwl_root_prefix, port_name),
                            schema_def_types=_get_schema_def_types(requirements),
                            format_graph=self.loading_context.loader.graph,
                            context=context,
                            check_type=False,
                            default_required_sf=False,
                            force_deep_listing=True))
                    # Add port as transformer input
                    transformer_step.add_input_port(port_name, port)
                    # If there are format dependencies, search for inputs in the port's input steps
                    for dep in format_deps:
                        port_found = False
                        for step in port.get_input_steps():
                            if dep in step.input_ports:
                                transformer_step.add_input_port(dep, step.get_input_port(dep))
                                port_found = True
                                break
                        if not port_found:
                            raise WorkflowDefinitionException("Cannot retrieve {} input port.".format(dep))
                    # Create an output port for the transformer
                    transformer_step.add_output_port(port_name, workflow.create_port())
                    # Create a schedule step and connect it to the local DeployStep
                    schedule_step = workflow.create_step(
                        cls=ScheduleStep,
                        name=posixpath.join(port_name + "-collector", "__schedule__"),
                        connector_port=deploy_step.get_output_port(),
                        input_directory=self.output_directory,
                        target=target)
                    # Add the port as an input of the schedule step
                    schedule_step.add_input_port(port_name, transformer_step.get_output_port())
                    # Add TransferStep to transfer the output in the output_dir
                    transfer_step = workflow.create_step(
                        cls=CWLTransferStep,
                        name=port_name + "-collector",
                        job_port=schedule_step.get_output_port(),
                        writable=True)
                    transfer_step.add_input_port(port_name, transformer_step.get_output_port())
                    transfer_step.add_output_port(port_name, workflow.create_port())
                    # Add the output port of the TransferStep to the workflow output ports
                    workflow.output_ports[port_name] = transfer_step.get_output_port().name
        # Return the final workflow object
        return workflow


class LinkMergeMethod(Enum):
    merge_nested = 1
    merge_flattened = 2


class ScatterPredecessor(object):
    __slots__ = ('step', 'scatter_step')

    def __init__(self,
                 step: str,
                 scatter_step: str):
        self.step: str = step
        self.scatter_step: str = scatter_step



from __future__ import annotations

import asyncio
import os
import sys
from abc import ABC
from typing import Any, MutableMapping, MutableSequence, Optional, Type

import aiosqlite
import pkg_resources
from cachetools import Cache, LRUCache

from streamflow.core import utils
from streamflow.core.asyncache import cachedmethod
from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import Target
from streamflow.core.persistence import Database, DependencyType
from streamflow.core.workflow import Port, Status, Step, Token


class CachedDatabase(Database, ABC):

    def __init__(self, context: StreamFlowContext):
        super().__init__(context)
        self.deployment_cache: Cache = LRUCache(maxsize=sys.maxsize)
        self.port_cache: Cache = LRUCache(maxsize=sys.maxsize)
        self.step_cache: Cache = LRUCache(maxsize=sys.maxsize)
        self.target_cache: Cache = LRUCache(maxsize=sys.maxsize)
        self.token_cache: Cache = LRUCache(maxsize=sys.maxsize)
        self.workflow_cache: Cache = LRUCache(maxsize=sys.maxsize)


class SqliteConnection(object):

    def __init__(self,
                 connection: str,
                 timeout: int,
                 init_db: bool):
        self.connection: str = connection
        self.timeout: int = timeout
        self.init_db: bool = init_db
        self._connection: Optional[aiosqlite.Connection] = None
        self.__row_factory = None

    async def __aenter__(self):
        if not self._connection:
            self._connection = await aiosqlite.connect(
                database=self.connection,
                timeout=self.timeout)
            if self.init_db:
                schema_path = pkg_resources.resource_filename(
                    __name__, os.path.join('schemas', 'sqlite.sql'))
                with open(schema_path, "r") as f:
                    async with self._connection.cursor() as cursor:
                        await cursor.execute("PRAGMA journal_mode = WAL")
                        await cursor.execute("PRAGMA wal_autocheckpoint = 10")
                        await cursor.executescript(f.read())
        self.__row_factory = self._connection.row_factory
        return self._connection

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._connection.row_factory = self.__row_factory

    async def close(self):
        if self._connection:
            await self._connection.close()


class SqliteDatabase(CachedDatabase):

    def __init__(self,
                 context: StreamFlowContext,
                 connection: str,
                 timeout: int = 20,
                 reset_db: bool = False):
        super().__init__(context)
        # If needed, reset the database
        if reset_db and os.path.isfile(connection):
            os.remove(connection)
        # Open connection to database
        os.makedirs(os.path.dirname(connection), exist_ok=True)
        self.connection: SqliteConnection = SqliteConnection(
            connection=connection,
            timeout=timeout,
            init_db=reset_db or not os.path.exists(connection))

    async def close(self):
        async with self.connection as db:
            await db.commit()
        await self.connection.close()

    @classmethod
    def get_schema(cls):
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'sqlite.json'))

    async def add_command(self,
                          step_id: int,
                          cmd: str) -> int:
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO command(step, cmd) "
                    "VALUES(:step, :cmd)", {
                        "step": step_id,
                        "cmd": cmd}) as cursor:
                return cursor.lastrowid

    async def add_dependency(self,
                             step: int,
                             port: int,
                             type: DependencyType,
                             name: str) -> None:
        async with self.connection as db:
            await db.execute(
                "INSERT OR IGNORE INTO dependency(step, port, type, name) "
                "VALUES(:step, :port, :type, :name)", {
                    "step": step,
                    "port": port,
                    "type": type.value,
                    "name": name})

    async def add_deployment(self,
                             name: str,
                             type: str,
                             config: str,
                             external: bool,
                             lazy: bool,
                             workdir: Optional[str]) -> int:
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO deployment(name, type, config, external, lazy, workdir) "
                    "VALUES (:name, :type, :config, :external, :lazy, :workdir)", {
                        "name": name,
                        "type": type,
                        "config": config,
                        "external": external,
                        "lazy": lazy,
                        "workdir": workdir}) as cursor:
                return cursor.lastrowid

    async def add_port(self,
                       name: str,
                       workflow_id: int,
                       type: Type[Port],
                       params: str) -> int:
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO port(name, workflow, type, params) "
                    "VALUES(:name, :workflow, :type, :params)", {
                        "name": name,
                        "workflow": workflow_id,
                        "type": utils.get_class_fullname(type),
                        "params": params}) as cursor:
                return cursor.lastrowid

    async def add_provenance(self,
                             inputs: MutableSequence[int],
                             token: int):
        provenance = [{'dependee': i, 'depender': token} for i in inputs]
        async with self.connection as db:
            await asyncio.gather(*(asyncio.create_task(db.execute(
                "INSERT OR IGNORE INTO provenance(dependee, depender) "
                "VALUES(:dependee, :depender)", prov)) for prov in provenance))

    async def add_step(self,
                       name: str,
                       workflow_id: int,
                       status: int,
                       type: Type[Step],
                       params: str) -> int:
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO step(name, workflow, status, type, params) "
                    "VALUES(:name, :workflow, :status, :type, :params)", {
                        "name": name,
                        "workflow": workflow_id,
                        "status": status,
                        "type": utils.get_class_fullname(type),
                        "params": params}) as cursor:
                return cursor.lastrowid

    async def add_target(self,
                         deployment: int,
                         type: Type[Target],
                         params: str,
                         locations: int = 1,
                         service: Optional[str] = None,
                         workdir: Optional[str] = None) -> int:
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO target(params, type, deployment, locations, service, workdir) "
                    "VALUES (:params, :type, :deployment, :locations, :service, :workdir)", {
                        "params": params,
                        "type": utils.get_class_fullname(type),
                        "deployment": deployment,
                        "locations": locations,
                        "service": service,
                        "workdir": workdir}) as cursor:
                return cursor.lastrowid

    async def add_token(self,
                        tag: str,
                        type: Type[Token],
                        value: Any,
                        port: Optional[int] = None):
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO token(port, type, tag, value) "
                    "VALUES(:port, :type, :tag, :value)", {
                        "port": port,
                        "type": utils.get_class_fullname(type),
                        "tag": tag,
                        "value": value}) as cursor:
                return cursor.lastrowid

    async def add_workflow(self,
                           name: str,
                           params: str,
                           status: int,
                           type: str) -> int:
        async with self.connection as db:
            async with db.execute(
                    "INSERT INTO workflow(name, params, status, type) "
                    "VALUES(:name, :params, :status, :type)", {
                        "name": name,
                        "params": params,
                        "status": status,
                        "type": type}) as cursor:
                return cursor.lastrowid

    @cachedmethod(lambda self: self.deployment_cache)
    async def get_deployment(self, deplyoment_id: int) -> MutableMapping[str, Any]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM deployment WHERE id = :id", {"id": deplyoment_id}) as cursor:
                return await cursor.fetchone()

    async def get_input_ports(self, step_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM dependency WHERE step = :step AND type = :type", {
                "step": step_id,
                "type": DependencyType.INPUT.value}) as cursor:
                return await cursor.fetchall()

    async def get_output_ports(self, step_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM dependency WHERE step = :step AND type = :type", {
                "step": step_id,
                "type": DependencyType.OUTPUT.value}) as cursor:
                return await cursor.fetchall()

    @cachedmethod(lambda self: self.port_cache)
    async def get_port(self, port_id: int) -> MutableMapping[str, Any]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM port WHERE id = :id", {"id": port_id}) as cursor:
                return await cursor.fetchone()

    async def get_port_tokens(self, port_id: int) -> MutableSequence[int]:
        async with self.connection as db:
            async with db.execute("SELECT id FROM token WHERE port = :port", {"port": port_id}) as cursor:
                return [row[0] for row in await cursor.fetchall()]

    async def get_report(self, workflow: str) -> MutableSequence[MutableMapping[str, Any]]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT c.id, s.name, c.start_time, c.end_time "
                                  "FROM step AS s, command AS c "
                                  "WHERE s.id = c.step "
                                  "AND s.workflow = (SELECT id FROM workflow WHERE name = :workflow)",
                                  {"workflow": workflow}) as cursor:
                return [dict(r) for r in await cursor.fetchall()]

    @cachedmethod(lambda self: self.step_cache)
    async def get_step(self, step_id: int) -> MutableMapping[str, Any]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM step WHERE id = :id", {"id": step_id}) as cursor:
                return await cursor.fetchone()

    @cachedmethod(lambda self: self.target_cache)
    async def get_target(self, target_id: int) -> MutableMapping[str, Any]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM target WHERE id = :id", {"id": target_id}) as cursor:
                return await cursor.fetchone()

    @cachedmethod(lambda self: self.token_cache)
    async def get_token(self, token_id: int) -> MutableMapping[str, Any]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM token WHERE id = :id", {"id": token_id}) as cursor:
                return await cursor.fetchone()

    async def get_workflow(self, workflow_id: int) -> MutableMapping[str, Any]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM workflow WHERE id = :id", {"id": workflow_id}) as cursor:
                return await cursor.fetchone()

    async def get_workflow_ports(self, workflow_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM port WHERE workflow = :workflow", {"workflow": workflow_id}) as cursor:
                return await cursor.fetchall()

    async def get_workflow_steps(self, workflow_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT * FROM step WHERE workflow = :workflow", {"workflow": workflow_id}) as cursor:
                return await cursor.fetchall()

    async def list_workflows(self) -> MutableSequence[MutableMapping[str, Any]]:
        async with self.connection as db:
            db.row_factory = aiosqlite.Row
            async with db.execute("SELECT name, status, type FROM workflow ORDER BY id DESC") as cursor:
                return [{'name': row[0], 'status': Status(row[1]).name, 'type': row[2]}
                        async for row in cursor]

    async def update_command(self,
                             command_id: int,
                             updates: MutableMapping[str, Any]) -> int:
        async with self.connection as db:
            await db.execute("UPDATE command SET {} WHERE id = :id".format(
                ", ".join(["{} = :{}".format(k, k) for k in updates])
            ), {**updates, **{"id": command_id}})
            return command_id

    async def update_deployment(self,
                                deployment_id: int,
                                updates: MutableMapping[str, Any]) -> int:
        async with self.connection as db:
            await db.execute("UPDATE deployment SET {} WHERE id = :id".format(
                ", ".join(["{} = :{}".format(k, k) for k in updates])
            ), {**updates, **{"id": deployment_id}})
            self.deployment_cache.pop(deployment_id, None)
            return deployment_id

    async def update_port(self,
                          port_id: int,
                          updates: MutableMapping[str, Any]) -> int:
        async with self.connection as db:
            await db.execute("UPDATE port SET {} WHERE id = :id".format(
                ", ".join(["{} = :{}".format(k, k) for k in updates])
            ), {**updates, **{"id": port_id}})
            self.port_cache.pop(port_id, None)
            return port_id

    async def update_step(self,
                          step_id: int,
                          updates: MutableMapping[str, Any]) -> int:
        async with self.connection as db:
            await db.execute("UPDATE step SET {} WHERE id = :id".format(
                ", ".join(["{} = :{}".format(k, k) for k in updates])
            ), {**updates, **{"id": step_id}})
            self.step_cache.pop(step_id, None)
            return step_id

    async def update_target(self,
                            target_id: int,
                            updates: MutableMapping[str, Any]) -> int:
        async with self.connection as db:
            await db.execute("UPDATE target SET {} WHERE id = :id".format(
                ", ".join(["{} = :{}".format(k, k) for k in updates])
            ), {**updates, **{"id": target_id}})
            self.target_cache.pop(target_id, None)
            return target_id

    async def update_workflow(self,
                              workflow_id: int,
                              updates: MutableMapping[str, Any]) -> int:
        async with self.connection as db:
            await db.execute("UPDATE workflow SET {} WHERE id = :id".format(
                ", ".join(["{} = :{}".format(k, k) for k in updates])),
                {**updates, **{"id": workflow_id}})
            self.workflow_cache.pop(workflow_id, None)
            return workflow_id



from __future__ import annotations

import asyncio
import io
import os
import posixpath
import shlex
import tarfile
import uuid
from abc import ABC
from pathlib import Path
from shutil import which
from typing import Any, Coroutine, MutableMapping, MutableSequence, Optional, Tuple, Union, cast

import pkg_resources
import yaml
from cachetools import Cache, TTLCache
from kubernetes_asyncio import client
from kubernetes_asyncio.client import ApiClient, Configuration, V1Container
from kubernetes_asyncio.config import ConfigException, load_incluster_config, load_kube_config
from kubernetes_asyncio.stream import WsApiClient, ws_client

from streamflow.core import utils
from streamflow.core.asyncache import cachedmethod
from streamflow.core.context import StreamFlowContext
from streamflow.core.data import StreamWrapperContext
from streamflow.core.deployment import Connector
from streamflow.core.exception import WorkflowExecutionException
from streamflow.core.scheduling import Location
from streamflow.data import aiotarstream
from streamflow.data.aiotarstream import BaseStreamWrapper
from streamflow.deployment.connector.base import BaseConnector
from streamflow.log_handler import logger

SERVICE_NAMESPACE_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"


def _check_helm_installed():
    if which("helm") is None:
        raise WorkflowExecutionException("Helm must be installed on the system to use the Helm connector.")


async def _get_helm_version():
    proc = await asyncio.create_subprocess_exec(
        *shlex.split("helm version --template '{{.Version}}'"),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.DEVNULL)
    stdout, _ = await proc.communicate()
    return stdout.decode().strip()


class KubernetesResponseWrapper(BaseStreamWrapper):

    async def read(self, size: Optional[int] = None):
        while not self.stream.closed:
            async for msg in self.stream:
                channel = msg.data[0]
                data = msg.data[1:]
                if data and channel == ws_client.STDOUT_CHANNEL:
                    return data

    async def write(self, data: Any):
        channel_prefix = bytes(chr(ws_client.STDIN_CHANNEL), "ascii")
        payload = channel_prefix + data
        await self.stream.send_bytes(payload)


class KubernetesResponseWrapperContext(StreamWrapperContext):

    def __init__(self,
                 coro: Coroutine):
        self.coro: Coroutine = coro
        self.response: Optional[KubernetesResponseWrapper] = None

    async def __aenter__(self):
        response = await self.coro
        self.response = KubernetesResponseWrapper(response)
        return self.response

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.response:
            await self.response.close()


class BaseKubernetesConnector(BaseConnector, ABC):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 inCluster: Optional[bool] = False,
                 kubeconfig: Optional[str] = os.path.join(str(Path.home()), ".kube", "config"),
                 namespace: Optional[str] = None,
                 locationsCacheSize: int = None,
                 locationsCacheTTL: int = None,
                 resourcesCacheSize: int = None,
                 resourcesCacheTTL: int = None,
                 transferBufferSize: int = (2 ** 25) - 1,
                 maxConcurrentConnections: int = 4096):
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            transferBufferSize=transferBufferSize)
        self.inCluster = inCluster
        self.kubeconfig = kubeconfig
        self.namespace = namespace
        cacheSize = locationsCacheSize
        if cacheSize is None:
            cacheSize = resourcesCacheSize
            if cacheSize is not None:
                logger.warn("The `resourcesCacheSize` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                            "Use `locationsCacheSize` instead.")
            else:
                cacheSize = 10
        cacheTTL = locationsCacheTTL
        if cacheTTL is None:
            cacheTTL = resourcesCacheTTL
            if cacheTTL is not None:
                logger.warn("The `resourcesCacheTTL` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                            "Use `locationsCacheTTL` instead.")
            else:
                cacheTTL = 10
        self.locationsCache: Cache = TTLCache(maxsize=cacheSize, ttl=cacheTTL)
        self.configuration: Optional[Configuration] = None
        self.client: Optional[client.CoreV1Api] = None
        self.client_ws: Optional[client.CoreV1Api] = None
        self.maxConcurrentConnections: int = maxConcurrentConnections

    def _configure_incluster_namespace(self):
        if self.namespace is None:
            if not os.path.isfile(SERVICE_NAMESPACE_FILENAME):
                raise ConfigException(
                    "Service namespace file does not exists.")

            with open(SERVICE_NAMESPACE_FILENAME) as f:
                self.namespace = f.read()
                if not self.namespace:
                    raise ConfigException("Namespace file exists but empty.")

    async def _copy_local_to_remote(self,
                                    src: str,
                                    dst: str,
                                    locations: MutableSequence[str],
                                    read_only: bool = False):
        effective_locations = await self._get_effective_locations(locations, dst)
        await super()._copy_local_to_remote(
            src=src,
            dst=dst,
            locations=effective_locations,
            read_only=read_only)

    async def _copy_local_to_remote_single(self,
                                           src: str,
                                           dst: str,
                                           location: str,
                                           read_only: bool = False) -> None:
        pod, container = location.split(':')
        command = ["tar", "xf", "-", "-C", "/"]
        # noinspection PyUnresolvedReferences
        response = await self.client_ws.connect_get_namespaced_pod_exec(
            name=pod,
            namespace=self.namespace or 'default',
            container=container,
            command=command,
            stderr=False,
            stdin=True,
            stdout=False,
            tty=False,
            _preload_content=False)
        try:
            async with aiotarstream.open(
                    stream=KubernetesResponseWrapper(response),
                    format=tarfile.GNU_FORMAT,
                    mode='w',
                    dereference=True,
                    copybufsize=self.transferBufferSize) as tar:
                await tar.add(src, arcname=dst)
        except tarfile.TarError as e:
            raise WorkflowExecutionException("Error copying {} to {} on location {}: {}".format(
                src, dst, location, str(e))) from e
        finally:
            await response.close()

    async def _copy_remote_to_local(self,
                                    src: str,
                                    dst: str,
                                    location: str,
                                    read_only: bool = False):
        pod, container = location.split(':')
        command = ["tar", "chf", "-", "-C", "/", posixpath.relpath(src, '/')]
        # noinspection PyUnresolvedReferences
        response = await self.client_ws.connect_get_namespaced_pod_exec(
            name=pod,
            namespace=self.namespace or 'default',
            container=container,
            command=command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False)
        try:
            async with aiotarstream.open(
                    stream=KubernetesResponseWrapper(response),
                    mode='r',
                    copybufsize=self.transferBufferSize) as tar:
                await utils.extract_tar_stream(tar, src, dst, self.transferBufferSize)
        except tarfile.TarError as e:
            raise WorkflowExecutionException("Error copying {} from location {} to {}: {}".format(
                src, location, dst, str(e))) from e
        finally:
            await response.close()

    async def _copy_remote_to_remote(self,
                                     src: str,
                                     dst: str,
                                     locations: MutableSequence[str],
                                     source_location: str,
                                     source_connector: Optional[Connector] = None,
                                     read_only: bool = False) -> None:
        source_connector = source_connector or self
        locations = await self._get_effective_locations(locations, dst)
        if source_connector == self and source_location in locations:
            if src != dst:
                command = ['/bin/cp', "-rf", src, dst]
                await self.run(source_location, command)
                locations.remove(source_location)
        if locations:
            # Get write command
            write_command = await utils.get_remote_to_remote_write_command(
                src_connector=source_connector,
                src_location=source_location,
                src=src,
                dst_connector=self,
                dst_locations=locations,
                dst=dst)
            async with source_connector._get_stream_reader(source_location, src) as reader:
                # Open a target response for each location
                writers = [KubernetesResponseWrapper(w) for w in await asyncio.gather(*(asyncio.create_task(
                    cast(Coroutine, self.client_ws.connect_get_namespaced_pod_exec(
                        name=location.split(':')[0],
                        namespace=self.namespace or 'default',
                        container=location.split(':')[1],
                        command=write_command,
                        stderr=True,
                        stdin=False,
                        stdout=True,
                        tty=False,
                        _preload_content=False))) for location in locations))]
                # Multiplex the reader output to all the writers
                while content := await reader.read(source_connector.transferBufferSize):
                    await asyncio.gather(*(asyncio.create_task(writer.write(content)) for writer in writers))

    async def _get_container(self, location: str) -> Tuple[str, V1Container]:
        pod_name, container_name = location.split(':')
        pod = await self.client.read_namespaced_pod(name=pod_name, namespace=self.namespace or 'default')
        for container in pod.spec.containers:
            if container.name == container_name:
                return container.name, container

    async def _get_configuration(self) -> Configuration:
        if self.configuration is None:
            self.configuration = Configuration()
            if self.inCluster:
                load_incluster_config(client_configuration=self.configuration)
                self._configure_incluster_namespace()
            else:
                await load_kube_config(config_file=self.kubeconfig, client_configuration=self.configuration)
        return self.configuration

    async def _get_effective_locations(self,
                                       locations: MutableSequence[str],
                                       dest_path: str,
                                       source_location: Optional[str] = None) -> MutableSequence[str]:
        # Get containers
        container_tasks = []
        for location in locations:
            container_tasks.append(asyncio.create_task(self._get_container(location)))
        containers = {k: v for (k, v) in await asyncio.gather(*container_tasks)}
        # Check if some locations share volume mounts to the same path
        common_paths = {}
        effective_locations = []
        for location in locations:
            container = containers[location.split(':')[1]]
            add_location = True
            for volume in container.volume_mounts:
                if dest_path.startswith(volume.mount_path):
                    path = ':'.join([volume.name, dest_path])
                    if path not in common_paths:
                        common_paths[path] = location
                    elif location == source_location:
                        effective_locations.remove(common_paths[path])
                        common_paths[path] = location
                    else:
                        add_location = False
                    break
            if add_location:
                effective_locations.append(location)
        return effective_locations

    def _get_run_command(self,
                         command: str,
                         location: str,
                         interactive: bool = False):
        pod, container = location.split(':')
        return (
            "kubectl "
            "{namespace}"
            "{kubeconfig}"
            "exec "
            "{pod} "
            "{interactive}"
            "{container}"
            "-- "
            "{command}"
        ).format(
            namespace=self.get_option("namespace", self.namespace),
            kubeconfig=self.get_option("kubeconfig", self.kubeconfig),
            pod=pod,
            interactive=self.get_option("i", interactive),
            container=self.get_option("container", container),
            command=command)

    def _get_stream_reader(self,
                           location: str,
                           src: str) -> StreamWrapperContext:
        pod, container = location.split(':')
        dirname, basename = posixpath.split(src)
        return KubernetesResponseWrapperContext(
            coro=cast(Coroutine, self.client_ws.connect_get_namespaced_pod_exec(
                name=pod,
                namespace=self.namespace or 'default',
                container=container,
                command=["tar", "chf", "-", "-C", dirname, basename],
                stderr=False,
                stdin=True,
                stdout=False,
                tty=False,
                _preload_content=False)))

    async def deploy(self, external: bool):
        # Init standard client
        configuration = await self._get_configuration()
        configuration.connection_pool_maxsize = self.maxConcurrentConnections
        self.client = client.CoreV1Api(api_client=ApiClient(configuration=configuration))
        # Init WebSocket client
        configuration = await self._get_configuration()
        configuration.connection_pool_maxsize = self.maxConcurrentConnections
        ws_api_client = WsApiClient(configuration=configuration, heartbeat=30)
        ws_api_client.set_default_header('Connection', 'upgrade,keep-alive')
        self.client_ws = client.CoreV1Api(api_client=ws_api_client)

    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        command = utils.create_command(
            command, environment, workdir, stdin, stdout, stderr)
        logger.debug("Executing command {command} on {location} {job}".format(
            command=command,
            location=location,
            job="for job {job}".format(job=job_name) if job_name else ""))
        command = utils.encode_command(command)
        pod, container = location.split(':')
        # noinspection PyUnresolvedReferences
        response = await self.client_ws.connect_get_namespaced_pod_exec(
            name=pod,
            namespace=self.namespace or 'default',
            container=container,
            command=["sh", "-c", "{command}".format(command=command)],
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=not capture_output)
        if capture_output:
            with io.StringIO() as out_buffer, io.StringIO() as err_buffer:
                while not response.closed:
                    async for msg in response:
                        data = msg.data.decode('utf-8', 'replace')
                        channel = ord(data[0])
                        data = data[1:]
                        if data and channel in [ws_client.STDOUT_CHANNEL, ws_client.STDERR_CHANNEL]:
                            out_buffer.write(data)
                        elif data and channel == ws_client.ERROR_CHANNEL:
                            err_buffer.write(data)
                await response.close()
                err = yaml.safe_load(err_buffer.getvalue())
                if err['status'] == "Success":
                    return out_buffer.getvalue(), 0
                else:
                    if 'code' in err:
                        return err['message'], int(err['code'])
                    else:
                        return err['message'], int(err['details']['causes'][0]['message'])

    async def undeploy(self, external: bool):
        if self.client is not None:
            await self.client.api_client.close()
            self.client = None
        if self.client_ws is not None:
            await self.client_ws.api_client.close()
            self.client_ws = None
        self.configuration = None


class Helm3Connector(BaseKubernetesConnector):
    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 chart: str,
                 debug: Optional[bool] = False,
                 kubeContext: Optional[str] = None,
                 kubeconfig: Optional[str] = None,
                 atomic: Optional[bool] = False,
                 caFile: Optional[str] = None,
                 certFile: Optional[str] = None,
                 depUp: Optional[bool] = False,
                 devel: Optional[bool] = False,
                 inCluster: Optional[bool] = False,
                 keepHistory: Optional[bool] = False,
                 keyFile: Optional[str] = None,
                 keyring: Optional[str] = None,
                 locationsCacheSize: int = None,
                 locationsCacheTTL: int = None,
                 nameTemplate: Optional[str] = None,
                 namespace: Optional[str] = None,
                 noHooks: Optional[bool] = False,
                 password: Optional[str] = None,
                 renderSubchartNotes: Optional[bool] = False,
                 repo: Optional[str] = None,
                 commandLineValues: Optional[MutableSequence[str]] = None,
                 fileValues: Optional[MutableSequence[str]] = None,
                 stringValues: Optional[MutableSequence[str]] = None,
                 registryConfig: Optional[str] = os.path.join(str(Path.home()), ".config/helm/registry.json"),
                 releaseName: Optional[str] = "release-%s" % str(uuid.uuid1()),
                 repositoryCache: Optional[str] = os.path.join(str(Path.home()), ".cache/helm/repository"),
                 repositoryConfig: Optional[str] = os.path.join(str(Path.home()), ".config/helm/repositories.yaml"),
                 resourcesCacheSize: int = None,
                 resourcesCacheTTL: int = None,
                 skipCrds: Optional[bool] = False,
                 timeout: Optional[str] = "1000m",
                 transferBufferSize: int = (32 << 20) - 1,
                 username: Optional[str] = None,
                 yamlValues: Optional[MutableSequence[str]] = None,
                 verify: Optional[bool] = False,
                 chartVersion: Optional[str] = None,
                 wait: Optional[bool] = True):
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            inCluster=inCluster,
            kubeconfig=kubeconfig,
            namespace=namespace,
            locationsCacheSize=locationsCacheSize,
            locationsCacheTTL=locationsCacheTTL,
            resourcesCacheSize=resourcesCacheSize,
            resourcesCacheTTL=resourcesCacheTTL,
            transferBufferSize=transferBufferSize)
        self.chart: str = os.path.join(context.config_dir, chart)
        self.debug: bool = debug
        self.kubeContext: Optional[str] = kubeContext
        self.atomic: bool = atomic
        self.caFile: Optional[str] = caFile
        self.certFile: Optional[str] = certFile
        self.depUp: bool = depUp
        self.devel: bool = devel
        self.keepHistory: bool = keepHistory
        self.keyFile: Optional[str] = keyFile
        self.keyring: Optional[str] = keyring
        self.nameTemplate: Optional[str] = nameTemplate
        self.noHooks: bool = noHooks
        self.password: Optional[str] = password
        self.renderSubchartNotes: bool = renderSubchartNotes
        self.repo: Optional[str] = repo
        self.commandLineValues: Optional[MutableSequence[str]] = commandLineValues
        self.fileValues: Optional[MutableSequence[str]] = fileValues
        self.stringValues: Optional[MutableSequence[str]] = stringValues
        self.skipCrds: bool = skipCrds
        self.registryConfig = registryConfig
        self.releaseName: str = releaseName
        self.repositoryCache = repositoryCache
        self.repositoryConfig = repositoryConfig
        self.timeout: Optional[str] = timeout
        self.username: Optional[str] = username
        self.yamlValues: Optional[MutableSequence[str]] = yamlValues
        self.verify: bool = verify
        self.chartVersion: Optional[str] = chartVersion
        self.wait: bool = wait

    def base_command(self):
        return (
            "helm "
            "{debug}"
            "{kubeContext}"
            "{kubeconfig}"
            "{namespace}"
            "{registryConfig}"
            "{repositoryCache}"
            "{repositoryConfig}"
        ).format(
            debug=self.get_option("debug", self.debug),
            kubeContext=self.get_option("kube-context", self.kubeContext),
            kubeconfig=self.get_option("kubeconfig", self.kubeconfig),
            namespace=self.get_option("namespace", self.namespace),
            registryConfig=self.get_option("registry-config", self.registryConfig),
            repositoryCache=self.get_option("repository-cache", self.repositoryCache),
            repositoryConfig=self.get_option("repository-config", self.repositoryConfig),
        )

    async def deploy(self, external: bool) -> None:
        # Create clients
        await super().deploy(external)
        if not external:
            # Check if Helm is installed
            _check_helm_installed()
            # Check correct version of Helm
            version = await _get_helm_version()
            if not version.startswith("v3"):
                raise WorkflowExecutionException(
                    "Helm {version} is not compatible with Helm3Connector".format(version=version))
            # Deploy Helm charts
            deploy_command = self.base_command() + "".join([
                "install "
                "{atomic}"
                "{caFile}"
                "{certFile}"
                "{depUp}"
                "{devel}"
                "{keyFile}"
                "{keyring}"
                "{nameTemplate}"
                "{noHooks}"
                "{password}"
                "{renderSubchartNotes}"
                "{repo}"
                "{commandLineValues}"
                "{fileValues}"
                "{stringValues}"
                "{skipCrds}"
                "{timeout}"
                "{username}"
                "{yamlValues}"
                "{verify}"
                "{chartVersion}"
                "{wait}"
                "{releaseName}"
                "{chart}"
            ]).format(
                atomic=self.get_option("atomic", self.atomic),
                caFile=self.get_option("ca-file", self.caFile),
                certFile=self.get_option("cert-file", self.certFile),
                depUp=self.get_option("dep-up", self.depUp),
                devel=self.get_option("devel", self.devel),
                keyFile=self.get_option("key-file", self.keyFile),
                keyring=self.get_option("keyring", self.keyring),
                nameTemplate=self.get_option("name-template", self.nameTemplate),
                namespace=self.get_option("namespace", self.namespace),
                noHooks=self.get_option("no-hooks", self.noHooks),
                password=self.get_option("password", self.password),
                renderSubchartNotes=self.get_option("render-subchart-notes", self.renderSubchartNotes),
                repo=self.get_option("repo", self.repo),
                commandLineValues=self.get_option("set", self.commandLineValues),
                fileValues=self.get_option("set-file", self.fileValues),
                stringValues=self.get_option("set-string", self.stringValues),
                skipCrds=self.get_option("skip-crds", self.skipCrds),
                timeout=self.get_option("timeout", self.timeout),
                username=self.get_option("username", self.username),
                yamlValues=self.get_option("values", self.yamlValues),
                verify=self.get_option("verify", self.verify),
                chartVersion=self.get_option("version", self.chartVersion),
                wait=self.get_option("wait", self.wait),
                releaseName="{releaseName} ".format(releaseName=self.releaseName),
                chart="\"{chart}\"".format(chart=self.chart)
            )
            logger.debug("Executing {command}".format(command=deploy_command))
            proc = await asyncio.create_subprocess_exec(
                *shlex.split(deploy_command),
                stderr=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL)
            await proc.wait()

    @cachedmethod(lambda self: self.locationsCache)
    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        pods = await self.client.list_namespaced_pod(
            namespace=self.namespace or 'default',
            label_selector="app.kubernetes.io/instance={}".format(self.releaseName),
            field_selector="status.phase=Running"
        )
        valid_targets = {}
        for pod in pods.items:
            # Check if pod is ready
            is_ready = True
            for condition in pod.status.conditions:
                if condition.status != 'True':
                    is_ready = False
                    break
            # Filter out not ready and Terminating locations
            if is_ready and pod.metadata.deletion_timestamp is None:
                for container in pod.spec.containers:
                    if not service or service == container.name:
                        location_name = pod.metadata.name + ':' + service
                        valid_targets[location_name] = Location(name=location_name, hostname=pod.status.pod_ip)
                        break
        return valid_targets

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'helm3.json'))

    async def undeploy(self, external: bool) -> None:
        if not external:
            # Undeploy
            undeploy_command = self.base_command() + (
                "uninstall "
                "{keepHistory}"
                "{noHooks}"
                "{timeout}"
                "{releaseName}"
            ).format(
                keepHistory=self.get_option("keep-history", self.keepHistory),
                noHooks=self.get_option("no-hooks", self.noHooks),
                timeout=self.get_option("timeout", self.timeout),
                releaseName=self.releaseName
            )
            logger.debug("Executing {command}".format(command=undeploy_command))
            proc = await asyncio.create_subprocess_exec(*shlex.split(undeploy_command))
            await proc.wait()
        # Close connections
        await super().undeploy(external)
